#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2009-2020 Authors of CryptoMiniSat, see AUTHORS file
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from __future__ import print_function
import optparse
import time
import pickle
import pandas as pd
import numpy as np
import sys


if __name__ == "__main__":
    usage = """usage: %prog [options] file1.dat [file2.dat ...]
Concatenates Pandas dataframes into a single file. The dataframes must have
the same columns."""
    parser = optparse.OptionParser(usage=usage)

    parser.add_option("--verbose", "-v", action="store_true", default=False,
                      dest="verbose", help="Print more output")
    parser.add_option("--out", "-o", default=None, type=str,
                      dest="out", help="Put combined output here")
    parser.add_option("--csv", default=None, type=str,
                      dest="csv", help="Dump CSV here")
    parser.add_option("--csvratio", default=1.0, type=float,
                      dest="csvratio", help="Ratio of data to dump to CSV")

    (options, args) = parser.parse_args()

    if options.out is None and options.csv is None:
        print("ERROR: you must either dump CSV or Pandas. Give either '--out' or '--csv' options")
        exit(-1)

    if options.csvratio < 0.0:
        print("ERROR: CSV ratio cannot be less than 0.0")
        exit(-1)

    if options.csvratio > 1.0:
        print("ERROR: CSV ratio cannot be more than 1.0")
        exit(-1)

    dfs = []
    for fname in args:
        print("----- Reading file %s -------" % fname)
        df = pd.read_pickle(fname)
        print("Read {num} datapoints from {file}".format(num=df.shape[0], file=fname))
        df["fname"] = df["fname"].astype("str")
        #df["cl.cur_restart_type"] = df["cl.cur_restart_type"].astype("str")
        #df["rdb0.cur_restart_type"] = df["rdb0.cur_restart_type"].astype("str")
        dfs.append(df)

    print("Concatenating dataframes...")
    df_full = pd.concat(dfs)
    print("Concated frame size: %d" % df_full.shape[0])

    if options.out is not None:
        fname = options.out
        print("Dumping {num} datapoint pandas data to: {fname}".format(num=df_full.shape[0], fname=fname))
        with open(fname, "wb") as f:
            pickle.dump(df_full, f)

    if options.csv is not None:
        import sklearn.model_selection

        fname = options.csv
        print("Dumpiing CSV ratio: %-3.2f" % options.csvratio)
        _, df_tmp = sklearn.model_selection.train_test_split(df_full, test_size=options.csvratio)
        print("Dumping {num} datapoint CSV data to: {fname}".format(num=df_tmp.shape[0], fname=fname))
        df_tmp.to_csv(fname, index=False, columns=sorted(list(df_tmp)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2022 Authors of CryptoMiniSat, see AUTHORS file
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.


# hard example
# ----------------
# ../scripts/create_random_bnn.py 300 3000 100 4 > test.bnn
# ./cryptominisat5  --occsimp 0 --scc 0 test.bnn --presimp 0 --varelim 0 --verb 1 -n1 --printsol 0

# very hard example
#-------------
# ../scripts/create_random_bnn.py 300 2000 100 7 > test.bnn
# ./cryptominisat5  --occsimp 0 --scc 0 test.bnn --presimp 0 --varelim 0 --verb 4 -n1 --printsol 0
# appmc counting this:
# c [appmc] [   21.46 ] bounded_sol_count looking for   73 solutions -- hashes active: 128
# c [appmc] [   39.18 ] bounded_sol_count looking for   73 solutions -- hashes active: 256
# c [appmc] [  117.86 ] bounded_sol_count looking for   73 solutions -- hashes active: 512

# more constraints than above
# --------------
# ../scripts/create_random_bnn.py 400 2000 100 9 > test.bnn
#c [appmc+arjun] Total time: 214.48
#c [appmc] Number of solutions is: 33*2**257
#s SATISFIABLE
#s mc 7642277889662868897955685010573401918315818987932277226604200544522266556235776
# BUG in appmc: test.bnn-appmcbug (also, bug-appmc in approxmc/build directory)

# kinda interesting one to run
# ---------------
# ../scripts/create_random_bnn.py 20 100 50 9 > test.bnn
# c Total time (this thread) : 41.60
# c [appmc+arjun] Total time: 41.59
# c [appmc] Number of solutions is: 52*2**59




import random
import sys

n = 100 # number of constraints
v = 1000 # number of variables
s = 30 # max size of constraints
seed = 1

if len(sys.argv) != 1:
    if len(sys.argv) != 5:
        print("ERROR: must give: n v s seed")
        exit(-1)

    n = int(sys.argv[1])
    v = int(sys.argv[2])
    s = int(sys.argv[3])
    seed = int(sys.argv[4])

    assert n > 10
    assert v > 0

random.seed(seed)
print("c n=%d v=%d seed=%d" % (n, v, seed))




for x in range(n):
    sz = random.randint(1,s)

    # get input + output
    ks = []
    while len(ks) < sz+1:
        x = random.randint(1,v)
        if x in ks or -x in ks:
            continue
        else:
            ks.append(x)

    # get cutoff
    cutoff = random.randint(0, sz)

    # get output
    output  = ks[len(ks)-1]
    ks = ks[:-1]

    pr = "b "
    for k in ks:
        if random.choice([True, False]):
            pr+="-"
        pr += "%d " % k
    pr += "0 "
    pr += "%d " % cutoff
    pr += "%d" % output
    print("%s" % pr)


# add some clauses
for x in range(n):
    sz = random.randint(1,8)
    ks = []
    while len(ks) < sz:
        x = random.randint(1, v)
        if x in ks or -x in ks:
            continue

        if random.choice([True, False]):
            ks.append(x)
        else:
            ks.append(-x)

    pr = ""
    for k in ks:
        pr += "%d " % k
    print("%s0" % pr)



#!/usr/bin/env python

# Copyright (c) 2017, Martin Horenovsky
# Copyright (C) 2009-2020 Authors of CryptoMiniSat, see AUTHORS file

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import sys
import hashlib

PY3 = sys.version_info.major == 3

input_name = sys.argv[1]
output_path = sys.argv[2]
last_dash = input_name.rfind("/")
if last_dash != -1:
    output_name = input_name[last_dash+1:].replace('.', '_')
else:
    output_name = input_name.replace('.', '_')


# In python 3, opening file as rb will return bytes and iteration is per byte
# In python 2, opening file as rb will return string and iteration is per char
# and char need to be converted to bytes.
# This function papers over the differences
def convert(c):
    if PY3:
        return c
    return ord(c)


with open(input_name, 'rb') as file:
    contents = file.read()


with open(output_path, 'w') as out:
    out.write('unsigned char {}[] = {{'.format(output_name))
    first = True
    for i, byte in enumerate(contents):
        if not first:
            out.write(', ')
        first = False
        if i % 12 == 0:
            out.write('\n  ')
        out.write('0x{:02x}'.format(convert(byte)))

    out.write(', 0x00')
    out.write('\n};\n')

    out.write('unsigned int {}_len = {};\n'.format(output_name, len(contents)))
    out.write('const char* {}_hash = "{}";\n'.format(output_name, hashlib.sha1(contents).hexdigest()))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2009-2020 Authors of CryptoMiniSat, see AUTHORS file
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import optparse
import random
import os

class PlainHelpFormatter(optparse.IndentedHelpFormatter):

    def format_description(self, description):
        if description:
            return description + "\n"
        else:
            return ""


usage = "usage: %prog [options] "
desc = """Generates an LPN problem
"""


def set_up_parser():
    parser = optparse.OptionParser(usage=usage, description=desc,
                                   formatter=PlainHelpFormatter())

    parser.add_option("--verbose", "-v", action="store_true", default=False,
                      dest="verbose", help="Print more output")

    # for fuzz-testing
    parser.add_option("--seed", "-s", dest="seed", default=None,
                      help="Genereate with this seed", type=int)
    parser.add_option("-n", dest="n", default=None, type=int,
                      help="Functiion width")
    parser.add_option("--pb", dest="pb", default=False, action="store_true",
                      help="Functiion width")
    parser.add_option("-m", dest="samples", default=None, type=int,
                      help="Number of samples")
    parser.add_option("--noise", dest="noise", default=0.1, type=float,
                      help="Ratio of noise")
    parser.add_option("--tolerance", "-t", dest="tolerance", default=None, type=int,
                      help="Tolerance for error. It is set AUTOMATICALLY if you don't set it here. It will be set to EXACTLY the actual error, so you will get a SAT instance.")
    #parser.add_option("--autotolerance", dest="autotolerance", action="store_true", default=False,
                      #help="Set tolerance automatically to PERFECT match. Default.")

    return parser



if __name__ == "__main__":
    # parse opts
    parser = set_up_parser()
    (opts, args) = parser.parse_args()

    if opts.seed is None or opts.n is None or opts.samples is None or opts.noise is None:
        print("ERROR: You MUST give all of: --seed, -n, -m")
        exit(-1)

    print("c Seed:          %3d" % opts.seed)
    print("c n:             %3d" % opts.n)
    print("c Samples:       %3d" % opts.samples)
    print("c Noise:           %-3.2f" % opts.noise)
    random.seed(opts.seed)

    # y[i] = set randomly [[output]]
    # x[i] = data that SAT solver can manipulate
    # val[x][y] = x[i]*y[i]


    fun = []
    for i in range(opts.n):
        fun.append(random.randint(0, 1))

    inputs = []
    for i in range(opts.samples):
        inp = []
        for i2 in range(opts.n):
            inp.append(random.randint(0, 1))
        inputs.append(inp)

    real_outputs = []
    for i in range(opts.samples):
        out = 0
        for a,b in zip(fun, inputs[i]):
            out ^= a*b
        real_outputs.append(out)

    num_incorrect_eqs = 0
    correct_eqs = []
    outputs = []
    for i in range(opts.samples):
        out = real_outputs[i]
        if random.random() < opts.noise:
            out = out ^ 1
            num_incorrect_eqs+=1
            correct_eqs.append(False)
        else:
            correct_eqs.append(True)
        outputs.append(out)

    print("c Tncorrect eqs: %3d" %num_incorrect_eqs)
    tolerance=None
    if opts.tolerance is None:
        tolerance = num_incorrect_eqs
        print("c -------------------- setting tolerance AUTOMATICALLY, since '--tolerance' was not set")
    else:
        tolerance = opts.tolerance
    print("c Tolerance:     %3d" % tolerance)

    # print
    print("c equations. FUN[i2]*INPUT[i][i2]")
    for i in range(opts.samples):
        toprint = ""
        for i2 in range(opts.n):
            toprint += "c %d" % fun[i2]
            toprint += "*"
            toprint += "%d" % inputs[i][i2]
            if i2 != opts.n-1:
                toprint += " + "
        toprint += " = %d" % outputs[i]
        toprint += "  -- correct: %s" % correct_eqs[i]
        print(toprint)


    # variable table (sequential):
    # n: function we need to figure out
    # inputs -- n+n+n...n exactly opts.samples times. Total: opts.samples*opts.n
    # outputs -- opts.samples
    # helper functions come here

    v = 1
    vars_fun = []
    for _ in range(opts.n):
        vars_fun.append(v)
        v+=1

    #vars_inputs = []
    #for _ in range(opts.samples):
        #tmp = []
        #for _ in range(opts.n):
            #tmp.append(v)
            #v+=1
        #vars_inputs.append(tmp)

    vars_noise = []
    for _ in range(opts.samples):
        vars_noise.append(v)
        v+=1

    ####################
    ####### Generate CNF
    ####################

    if opts.pb:
        print("* #variable= %d #constraint= 1" % (opts.n+opts.samples))

    # compute outputs
    for i in range(opts.samples):
        vs = []
        for i2 in range(opts.n):
            # v = inputs[i][i2] * fun[i]
            if inputs[i][i2] == 1:
                vs.append(vars_fun[i2])

        out = "x "
        out_pb = "* xor "
        for x in vs:
            out += "%d " % x
            out_pb += "x%d " % x

        out_pb += "x%d " % vars_noise[i]
        if outputs[i]:
            out+="%d " % vars_noise[i]
            out_pb+=" 1"
        else:
            out+="-%d " % vars_noise[i]
            out_pb+=" 0"

        out +="0"
        if opts.pb:
            print(out_pb)
        else:
            print(out)

    # get noise
    print("c Num equations: ", opts.samples)
    print("c Incorrect equations: ", num_incorrect_eqs)
    print("c Tolerance: ", tolerance)
    if num_incorrect_eqs > tolerance:
        print("c this will be UNSAT for sure. Tolerance is smaller than the number of Incorrect euqations.")

    if opts.pb:
        out = ""
        for i in range(opts.samples):
            out +="-1 x%d " % vars_noise[i]
        out += " >= %d" % (-tolerance)
        print(out)
    else:
        out = "b "
        for i in range(opts.samples):
            out +="-%d " % vars_noise[i]
        out += "0 %d" % (opts.samples-tolerance)
        print(out)

    # print correct output
    out = "c correct output is: "
    if num_incorrect_eqs <= tolerance:
        out += "SAT "
        for i in range(opts.n):
            val = fun[i]
            if val:
                out+= "%d " % vars_fun[i]
            else:
                out+= "-%d " % vars_fun[i]
    else:
        out += "UNSAT "
    print(out)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2009-2020 Authors of CryptoMiniSat, see AUTHORS file
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from __future__ import print_function
import optparse
import time
import pickle
import re
import pandas as pd
import numpy as np
import sys
import helper


class QueryAddIdxes (helper.QueryHelper):
    def __init__(self, dbfname):
        super(QueryAddIdxes, self).__init__(dbfname)

    def measure_size(self):
        self.c.execute("select count() from reduceDB")
        rows = self.c.fetchall()
        rdb_rows = rows[0][0]
        print("We have %d lines of RDB" % (rdb_rows))

        self.c.execute("select count() from clause_stats")
        rows = self.c.fetchall()
        clss_rows = rows[0][0]
        print("We have %d lines of clause_stats" % (clss_rows))

    def create_indexes(self):
        t = time.time()
        print("Recreating indexes...")
        queries = """
        create index `idxclid33` on `sum_cl_use` (`clauseID`, `last_confl_used`);
        ---
        create index `idxclid1` on `clause_stats` (`clauseID`, conflicts, latest_satzilla_feature_calc);
        create index `idxclid1-2` on `clause_stats` (`clauseID`);
        create index `idxclid2` on `clause_stats` (clauseID, conflicts, latest_satzilla_feature_calc);
        create index `idxclid5` on `tags` ( `name`);
        ---
        create index `idxclid6` on `reduceDB` (`clauseID`, conflicts);
        create index `idxclid6-9` on `reduceDB` (`conflicts`);
        create index `idxclid9` on `reduceDB_common` (`conflicts`, `latest_satzilla_feature_calc`);
        create index `idxclid9-2` on `reduceDB_common` (`conflicts`, `latest_satzilla_feature_calc`);
        create index `idxclid9-3` on `reduceDB_common` (`conflicts`);
        create index `idxclid9-4` on `reduceDB_common` (`latest_satzilla_feature_calc`);
        create index `idxclid6-2` on `reduceDB` (`clauseID`, `dump_no`);
        create index `idxclid6-3` on `reduceDB` (`clauseID`, `conflicts`, `dump_no`);
        create index `idxclid6-4` on `reduceDB` (`clauseID`, `conflicts`)
        ---
        create index `idxclid7` on `satzilla_features` (`latest_satzilla_feature_calc`);
        ---
        create index `idxclidUCLS-1` on `used_clauses` ( `clauseID`, `used_at`);
        create index `idxclidUCLS-2` on `used_clauses` ( `used_at`);
        ---
        create index `idxcl_last_in_solver-1` on `cl_last_in_solver` ( `clauseID`, `conflicts`);
        ---
        create index `used_later_percentiles_idx3` on `used_later_percentiles` (`type_of_dat`, `percentile_descr`, `percentile`, `val`);
        create index `used_later_percentiles_idx2` on `used_later_percentiles` (`type_of_dat`, `percentile_descr`, `val`);
        """
        for l in queries.split('\n'):
            t2 = time.time()

            if options.verbose:
                print("Creating index: ", l)
            self.c.execute(l)
            if options.verbose:
                print("Index creation T: %-3.2f s" % (time.time() - t2))

        print("indexes created T: %-3.2f s" % (time.time() - t))


class QueryCls (helper.QueryHelper):
    def __init__(self, dbfname, tier, table):
        super(QueryCls, self).__init__(dbfname)
        self.fill_sql_query(tier, table=table)

    def fill_sql_query(self, tier, table):
        # sum_cl_use
        self.sum_cl_use = helper.query_fragment(
            "sum_cl_use", [], "sum_cl_use", options.verbose, self.c)

        # RDB data
        not_cols = [
            "reduceDB_called"
            , "clauseID"
            , "in_xor"
            , "locked"
            , "conflicts"
            , "activity_rel"]
        self.rdb0_dat = helper.query_fragment(
            "reduceDB", not_cols, "rdb0", options.verbose, self.c)

        # reduceDB_common data
        not_cols = [
            "reduceDB_called"
            , "simplifications"
            , "restarts"
            #, "conflicts"
            , "latest_satzilla_feature_calc"
            , "runtime"
            ]
        self.rdb0_common_dat = helper.query_fragment(
            "reduceDB_common", not_cols, "rdb0_common", options.verbose, self.c)

        # clause data
        not_cols = [
            "simplifications"
            , "restarts"
            , "prev_restart"
            , "antecedents_long_red_age_max"
            , "antecedents_long_red_age_min"
            , "latest_satzilla_feature_calc"
            , "clauseID"]
        self.clause_dat = helper.query_fragment(
            "clause_stats", not_cols, "cl", options.verbose, self.c)

        # satzilla data
        not_cols = [
            "simplifications"
            , "restarts"
            , "conflicts"
            , "latest_satzilla_feature_calc"
            , "irred_glue_distr_mean"
            , "irred_glue_distr_var"]
        self.satzfeat_dat = helper.query_fragment(
            "satzilla_features", not_cols, "szfeat", options.verbose, self.c)

        self.common_limits = """
        order by random()
        limit {limit}
        """

        q_time_base="""
        join {table}_{tier} on
            {table}_{tier}.clauseID = rdb0.clauseID
            and {table}_{tier}.rdb0conflicts = rdb0.conflicts
        """

        q_columns_base="""
            , {table}_{tier}.used_later as `x.{table}_{tier}`
            , {table}_{tier}.percentile_fit as `x.{table}_{tier}_topperc`
            """

        q_time = q_time_base.format(tier=tier, table=table)
        q_columns = q_columns_base.format(tier=tier, table=table)

        # final big query
        self.q_select = """
        SELECT
        tags.val as `fname`
        {clause_dat}
        {rdb0_dat}
        {rdb0_common_dat}
        {sum_cl_use}
        , (rdb0.conflicts - rdb0.introduced_at_conflict) as `cl.time_inside_solver`
        , (sum_cl_use.last_confl_used - rdb0.introduced_at_conflict) as `x.a_lifetime`
        {q_columns}
        , sum_cl_use.num_used as `x.sum_cl_use`


        FROM
        reduceDB as rdb0

        -- this is DELIBERATEY left-join: this way, clauses that as ternary
        -- resolvents or otherwise generated during in-processing
        -- can still be used
        left join clause_stats as cl on
            cl.clauseID = rdb0.clauseID

        join reduceDB_common as rdb0_common on
            rdb0_common.conflicts = rdb0.conflicts

        join sum_cl_use on
            sum_cl_use.clauseID = rdb0.clauseID

        {q_time}

        join cl_last_in_solver on
            cl_last_in_solver.clauseID = rdb0.clauseID

        , tags

        WHERE
        (cl.clauseID != 0 OR cl.clauseID is NULL)
        and tags.name = "filename"

        -- to avoid missing clauses and their missing data to affect results
        and rdb0.conflicts + {del_at_least} <= cl_last_in_solver.conflicts
        """

        self.myformat = {
            "limit": 1000*1000*1000,
            "clause_dat": self.clause_dat,
            "satzfeat_dat_cur": self.satzfeat_dat.replace("szfeat.", "szfeat_cur."),
            "rdb0_dat": self.rdb0_dat,
            "sum_cl_use": self.sum_cl_use,
            "rdb0_common_dat": self.rdb0_common_dat,
            "q_time": q_time,
            "q_columns": q_columns
        }

    def get_used_later_percentiles(self, name, table):
        cur = self.conn.cursor()
        q = """
        select
            `type_of_dat`,
            `percentile_descr`,
            `percentile`,
            `val`
        from {table}_percentiles
        where `type_of_dat` = '{name}'
        """.format(name=name, table=table)
        cur.execute(q)
        rows = cur.fetchall()
        lookup = {}
        for row in rows:
            mystr = "%s_%s_perc" % (row[1], row[2])
            print("table: {table} type: {t}, perc_desc: {perc_desc}, perc: {perc}, val: {val}".format(
                table=table,
                t=row[0],
                perc_desc=row[1],
                perc=row[2],
                val=row[3]))
            lookup[mystr] = row[3]
        #print("perc lookup:", lookup)

        return lookup

    def get_one_data(self, tier, table):
        perc = self.get_used_later_percentiles(tier, table)
        self.myformat["del_at_least"] = getattr(options, tier)

        # when del_at_least is over 2 million, then we need to make this smaller
        #    or we will delete all data
        if self.myformat["del_at_least"] > 2*1000*1000:
            self.myformat["del_at_least"] = 100*1000

        # Make sure these stratas are equally represented
        t = time.time()
        dfs = self.run_stratified_queries(
            limit=options.limit, perc=perc, tier=tier, table=table)
        data = pd.concat(dfs)
        print("** Queries finished. Total size: %s  -- T: %-3.2f" % (
            data.shape, time.time() - t))
        return data

    # perc == percentile distribution
    # limit == MAX in total
    # tier == forever/long/short
    def run_stratified_queries(self, limit, perc, tier, table):
        dfs = []
        # NOTE: these are NON-ZERO percentages, but we replace 100 with "0", so the LAST chunk contains ALL, including 0, which is a large part of the data
        for beg_perc, end_perc in [(0.0, options.cut1), (options.cut1, options.cut2), (options.cut2, 100.0)]:
            beg = perc["top_non_zero_{perc}_perc".format(perc=beg_perc)]
            if end_perc == 100.0:
                end = 0.0
            else:
                end = perc["top_non_zero_{perc}_perc".format(perc=end_perc)]
            what_to_strata = "`x.{table}_{tier}`".format(tier=tier, table=table)

            print("Limit is {limit} value strata perc: ({a}, {b}) translates to value strata ({beg}, {end})".format(
                limit=limit, a=beg_perc, b=end_perc, beg=beg, end=end))

            # create one query for (beg,end) with different dump numbers
            df, weighted_sizes = self.query_strata_per_dumpno(
                str(self.q_select),
                limit,
                what_to_strata=what_to_strata,
                strata=(beg,end))
            dfs.append(df)
        return dfs

    def query_strata_per_dumpno(self, q, limit, what_to_strata, strata):
        print("* Getting one set of data with limit %s" % limit)
        weighted_size = []
        df_parts = []

        def one_part(mult, dump_no_filter=""):
            self.myformat["limit"] = int(limit*mult)
            df_parts.append(self.one_query(q + dump_no_filter, what_to_strata, strata))
            print("--> Num rows for strata %s -- '%s': %s" % (strata, dump_no_filter, df_parts[-1].shape[0]))

            ws = df_parts[-1].shape[0]/mult
            print("--> The weight was %f so weighted size is: %d" % (mult, int(ws)))
            weighted_size.append(ws)

        one_part(1/4.0, dump_no_filter=" and rdb0.dump_no = 1 ")
        one_part(1/4.0, dump_no_filter=" and rdb0.dump_no = 2 ")
        one_part(1/4.0, dump_no_filter=" and rdb0.dump_no > 2 ")
        one_part(1/4.0, dump_no_filter=" and rdb0.dump_no > 20 ")

        df = pd.concat(df_parts)
        print("-> size of all dump_no-s, strata {strata} data: {size}".format(
            strata=strata, size=df.shape))

        return df, weighted_size

    def one_query(self, q, what_to_strata, strata):
        q = q.format(**self.myformat)

        if strata[1] == 0.0:
            my_less_equal = ">="
        else:
            my_less_equal = ">"

        q = """
        select * from ( {q} )
        where
        {what_to_strata} <= {beg}
        and {what_to_strata} {my_less_equal} {end}""".format(
            q=q,
            what_to_strata=what_to_strata,
            beg=strata[0],
            end=strata[1],
            my_less_equal=my_less_equal)

        q += self.common_limits
        q = q.format(**self.myformat)

        t = time.time()
        sys.stdout.write("-> Running query for {} stratas {}...".format(what_to_strata, strata))
        sys.stdout.flush()
        if options.dump_sql:
            print("query:", q)
        df = pd.read_sql_query(q, self.conn)
        print("T: %-3.1f" % (time.time() - t))
        return df


def dump_dataframe(df, name):
    if options.dump_csv:
        fname = "%s.csv" % name
        print("Dumping CSV data to:", fname)
        df.to_csv(fname, index=False, columns=sorted(list(df)))

    fname = "%s.dat" % name
    print("Dumping pandas data to:", fname)
    with open(fname, "wb") as f:
        pickle.dump(df, f)


def one_database(dbfname):
    with QueryAddIdxes(dbfname) as q:
        q.measure_size()
        helper.drop_idxs(q.c)
        q.create_indexes()

    with helper.QueryFill(dbfname) as q:
        q.delete_and_create_used_laters()
        for tier in ["short", "long", "forever"]:
            for table in ["used_later", "used_later_anc"]:
                q.fill_used_later_X(tier, duration=getattr(options, tier), table=table)
                q.fill_used_later_X_perc_fit(tier, table=table)

    print("Using sqlite3 DB file %s" % dbfname)
    for tier in ["short", "long", "forever"]:
        for table in ["used_later", "used_later_anc"]:
            print("------> Doing tier {tier} table {table}".format(
                tier=tier,table=table))

            with QueryCls(dbfname, tier, table) as q:
                df = q.get_one_data(tier, table)

            if options.verbose:
                print("Describing----")
                dat = df.describe()
                print(dat)
                print("Describe done.---")
                print("Features: ", df.columns.values.flatten().tolist())

            if options.verbose:
                print("Describing post-transform ----")
                print(df.describe())
                print("Describe done.---")
                print("Features: ", df.columns.values.flatten().tolist())

            cleanname = re.sub(r'\.cnf.gz.sqlite$', '', dbfname)
            cleanname = re.sub(r'\.db$', '', dbfname)
            cleanname = re.sub(r'\.sqlitedb$', '', dbfname)
            cleanname = "{cleanname}-cldata-{table}-{tier}-cut1-{cut1}-cut2-{cut2}-limit-{limit}".format(
                cleanname=cleanname,
                cut1=options.cut1,
                cut2=options.cut2,
                limit=options.limit,
                tier=tier,
                table=table)

            dump_dataframe(df, cleanname)


if __name__ == "__main__":
    usage = "usage: %prog [options] file1.sqlite [file2.sqlite ...]"
    parser = optparse.OptionParser(usage=usage)

    # verbosity level
    parser.add_option("--verbose", "-v", action="store_true", default=False,
                      dest="verbose", help="Print more output")
    parser.add_option("--sql", action="store_true", default=False,
                      dest="dump_sql", help="Dump SQL queries")
    parser.add_option("--csv", action="store_true", default=False,
                      dest="dump_csv", help="Dump CSV (for weka)")

    # limits
    parser.add_option("--limit", default=20000, type=int,
                      dest="limit", help="Max number of samples to take from each strata (for each table/tier)")
    parser.add_option("--cut1", default=5.0, type=float,
                      dest="cut1", help="Where to cut the distrib. Default: %default")
    parser.add_option("--cut2", default=30.0, type=float,
                      dest="cut2", help="Where to cut the distrib. Default: %default")

    # debugging is faster with this
    parser.add_option("--noind", action="store_true", default=False,
                      dest="no_recreate_indexes",
                      help="Don't recreate indexes")

    # lengths of short/long
    parser.add_option("--short", default=10000, type=int,
                      dest="short", help="Short duration. Default: %default")
    parser.add_option("--long", default=30*1000, type=int,
                      dest="long", help="Long duration. Default: %default")
    parser.add_option("--forever", default=120*1000, type=int,
                      dest="forever", help="Long duration. Default: %default")

    (options, args) = parser.parse_args()

    if len(args) != 1:
        print("ERROR: You must give exactly one file")
        exit(-1)

    np.random.seed(2097483)
    one_database(args[0])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2009-2020 Authors of CryptoMiniSat, see AUTHORS file
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from __future__ import print_function
import sqlite3
import optparse
import time
import os.path
import helper


class QueryDatRem(helper.QueryHelper):
    def __init__(self, dbfname):
        super(QueryDatRem, self).__init__(dbfname)

    def create_percentiles_table(self):
        for table in ["used_later", "used_later_anc"]:
            # drop table
            q_drop = """
            DROP TABLE IF EXISTS `{table}_percentiles`;
            """
            self.c.execute(q_drop.format(table=table))

            # Create percentiles table
            q_create = """
            create table `{table}_percentiles` (
                `type_of_dat` string NOT NULL,
                `percentile_descr` string NOT NULL,
                `percentile` float DEFAULT NULL,
                `val` float NOT NULL
            );"""
            self.c.execute(q_create.format(table=table))

            idxs = """
            create index `{table}_percentiles_idx3` on `{table}_percentiles` (`type_of_dat`, `percentile_descr`, `percentile`, `val`);
            create index `{table}_percentiles_idx2` on `{table}_percentiles` (`type_of_dat`, `percentile_descr`, `val`);"""
            for q in idxs.split("\n"):
                self.c.execute(q.format(table=table))


    # "tier" here is "short", "long", or "forever"
    def get_all_percentile_X(self, tier):
        t = time.time()
        for table in ["used_later", "used_later_anc"]:
            print("Calculating percentiles now for table {table} and tier {tier} ...".format(
                tier=tier, table=table))

            q2 = """
            insert into {table}_percentiles (type_of_dat, percentile_descr, percentile, val)
            {q}
            """

            q = "select '{tier}', 'avg', NULL, avg(used_later) from {table}_{tier};".format(
                tier=tier, table=table)
            self.c.execute(q2.format(tier=tier, table=table, q=q))

            q = """
            SELECT
            '{tier}', 'top_non_zero', {perc}, used_later
            FROM {table}_{tier}
            WHERE used_later>0
            ORDER BY used_later ASC
            LIMIT 1
            OFFSET round((SELECT
             COUNT(*)
            FROM {table}_{tier}
            WHERE used_later>0) * ((100-{perc}) / 100.0)) - 1;
            """
            for perc in list(range(0,30,1))+list(range(30,100, 5)):
                myq = q.format(tier=tier, table=table, perc=perc)
                self.c.execute(q2.format(tier=tier, table=table, q=myq))
            # the 100% perecentile is not 0 (remember, this is "non-zero"), but let's cheat and add it in
            self.c.execute(q2.format(
                tier=tier, table=table, q="select '{tier}', 'top_non_zero', 100.0, 0.0;".format(
                    tier=tier, table=table)))

            q = """
            SELECT
            '{tier}', 'top_also_zero', {perc}, used_later
            FROM {table}_{tier}
            ORDER BY used_later ASC
            LIMIT 1
            OFFSET round((SELECT
             COUNT(*)
            FROM {table}_{tier}) * ((100.0-{perc}) / 100.0)) - 1;
            """
            for perc in range(0,100, 10):
                myq = q.format(tier=tier, table=table, perc=perc)
                self.c.execute(q2.format(tier=tier, table=table, q=myq))
            self.c.execute(
                q2.format(tier=tier, table=table,
                          q="select '{tier}', 'top_also_zero', 100.0, 0.0;".format(tier=tier, table=table)))

            print("Calculated percentiles/averages, T:", time.time()-t)

    def print_percentiles(self):
        q_check = "select * from used_later_percentiles"
        cur = self.conn.cursor()
        cur.execute(q_check)
        rows = cur.fetchall()
        print("Percentiles/average for used_later_percentiles:")
        for row in rows:
            print(" -> %s %s -- %s : %s" %(row[0], row[1], row[2], row[3]))


    def create_indexes1(self):
        print("Recreating indexes...")
        t = time.time()
        queries = """
        create index `idxclid31` on `clause_stats` (`clauseID`);
        create index `idxclid32` on `reduceDB` (`clauseID`);
        create index `idxclid33` on `sum_cl_use` (`clauseID`);
        create index `idxclid34` on `used_clauses` (`clauseID`);
        create index `idxclid44` on `restart_dat_for_cl` (`clauseID`);
        create index `idxclid35` on `var_data_fintime` (`var`, `sumConflicts_at_picktime`);
        create index `idxclid36` on `var_data_picktime` (`var`, `sumConflicts_at_picktime`);
        create index `idxclid37` on `dec_var_clid` (`var`, `sumConflicts_at_picktime`);
        create index `idxclid40` on `restart_dat_for_var` (`conflicts`);
        """

        for q in queries.split("\n"):
            self.c.execute(q)

        print("Created indexes needed T: %-3.2f s"% (time.time() - t))

    def recreate_used_ID_table(self):
        q = """
        DROP TABLE IF EXISTS `used_cl_ids`;
        """
        self.c.execute(q)

        q = """
        CREATE TABLE `used_cl_ids` (
          `clauseID` int(20) NOT NULL
        );
        """
        self.c.execute(q)

        q = """
        create index `idxclid30` on `used_cl_ids` (`clauseID`);
        """
        self.c.execute(q)

    def remove_too_many_vardata(self):
        t = time.time()
        q = """
        select count()
        from var_data_picktime
        """
        ret = self.c.execute(q)
        rows = self.c.fetchall()
        assert len(rows) == 1
        num_vardata = rows[0][0]
        print("Current number of elements in var_data: %d" % num_vardata)

        if num_vardata < options.goal_vardata:
            print("Not too many in var_data, skipping removal.")
            return

        q = """
        DROP TABLE IF EXISTS `used_vardat`;
        """
        self.c.execute(q)

        q = """
        CREATE TABLE `used_vardat` (
          `var` bigint(20) NOT NULL
          , `sumConflicts_at_picktime` bigint(20) NOT NULL
        );
        """
        self.c.execute(q)

        q = """
        create index `idxclidxx` on `used_vardat`
        (`var`, `sumConflicts_at_picktime`);
        """
        self.c.execute(q)

        q = """
        insert into `used_vardat`
        SELECT
        var, sumConflicts_at_picktime
        FROM var_data_picktime
        order by random()
        limit {limit}
        """.format(limit=options.goal_vardata)
        self.c.execute(q)
        print("Added {limit} to `used_vardat`".format(limit=options.goal_vardata))
        print("--> T: %-3.2f s"% (time.time() - t))

        t = time.time()
        del_from = ["var_data_picktime", "var_data_fintime", "dec_var_clid"]
        for table in del_from:

            q = """
            DROP TABLE IF EXISTS `myrows`;
            """
            self.c.execute(q)

            q = """
            CREATE TABLE `myrows` (
              `myrowid` bigint(20) NOT NULL
            );
            """
            self.c.execute(q)

            q = """
            INSERT INTO `myrows`
            SELECT `rowid`
            FROM `{table}` WHERE (`var`, `sumConflicts_at_picktime`)
            in (SELECT `var`, `sumConflicts_at_picktime` from `used_vardat`);
            """
            self.c.execute(q.format(table=table))

            q = """
            create index `myidx111` on `myrows` (`myrowid`);
            """

            q = """
            DELETE FROM `{table}` WHERE (rowid) NOT IN
            (SELECT `myrowid` from `myrows` );"""
            self.c.execute(q.format(table=table))
            print("Deleted unused data from %s" % table)

        # cleanup
        q = """
        DROP TABLE IF EXISTS `myrows`;
        """
        self.c.execute(q)

        # sample restart_dat_for_var
        q = """
        DELETE FROM restart_dat_for_var WHERE `conflicts` NOT IN
        (SELECT `sumConflicts_at_picktime` from `used_vardat` group by sumConflicts_at_picktime);"""
        self.c.execute(q)
        print("Deleted unused data from restart_dat_for_var")

        # cleanup
        q = """
        DROP TABLE IF EXISTS `used_vardat`;
        """
        self.c.execute(q)
        print("Cleaned up var_data_x & restart_dat_for_var tables T: %-3.2f s"
              % (time.time() - t))

    def insert_into_used_cls_ids_from_clstats(self, min_used, limit, table):
        min_used = int(min_used)

        t = time.time()
        val = int()
        q = """
        insert into used_cl_ids
        select
        clauseID from {table}
        where
        num_used >= {min_used}
        order by random() limit {limit}
        """.format(
            min_used=min_used,
            limit=int(limit),
            table=table)

        self.c.execute(q)
        print("Added num_used >= %d from sum_cl_use to used_cls_ids T: %-3.2f s"
              % (min_used, time.time() - t))

    # inserts ratio that's slanted towards >=1 use
    def fill_used_cl_ids_table(self, fair, limit):
        t = time.time()
        for table in ["sum_cl_use"]:
            if not fair:
                self.insert_into_used_cls_ids_from_clstats(min_used=100000, limit=limit/20, table=table)
                self.insert_into_used_cls_ids_from_clstats(min_used=50000, limit=limit/10, table=table)
                self.insert_into_used_cls_ids_from_clstats(min_used=10000, limit=limit/10, table=table)
                self.insert_into_used_cls_ids_from_clstats(min_used=1000, limit=limit/10, table=table)
                self.insert_into_used_cls_ids_from_clstats(min_used=100, limit=limit/10, table=table)
                self.insert_into_used_cls_ids_from_clstats(min_used=30, limit=limit/5, table=table)
                self.insert_into_used_cls_ids_from_clstats(min_used=20, limit=limit/4, table=table)
                self.insert_into_used_cls_ids_from_clstats(min_used=5, limit=limit/3, table=table)
                self.insert_into_used_cls_ids_from_clstats(min_used=1, limit=limit/2, table=table)
            self.insert_into_used_cls_ids_from_clstats(min_used=0, limit=limit/2, table=table)

        q = """
        select count()
        from used_cl_ids, sum_cl_use
        where
        used_cl_ids.clauseID = sum_cl_use.clauseID
        and sum_cl_use.num_used > 0
        """
        ret = self.c.execute(q)
        rows = self.c.fetchall()
        assert len(rows) == 1
        good_ids = rows[0][0]

        q = """
        select count()
        from used_cl_ids, sum_cl_use
        where
        used_cl_ids.clauseID = sum_cl_use.clauseID
        and sum_cl_use.num_used = 0
        """
        ret = self.c.execute(q)
        rows = self.c.fetchall()
        assert len(rows) == 1
        bad_ids = rows[0][0]

        print("IDs in used_cl_ids that are 'good' (sum_cl_use.num_used > 0) : %d" % good_ids)
        print("IDs in used_cl_ids that are 'bad'  (sum_cl_use.num_used = 0) : %d" % bad_ids)
        print("   T: %-3.2f s" % (time.time() - t))

    def print_idxs(self):

        q = """
        SELECT * FROM sqlite_master WHERE type == 'index'
        """
        self.c.execute(q)
        rows = self.c.fetchall()
        queries = ""

        if options.verbose:
            print("Using indexes: ")
            for row in rows:
                print("-> index:", row)

    def create_used_clauses_red(self):
        t = time.time()
        q = """
        CREATE TABLE used_clauses_red AS
        SELECT * FROM used_clauses WHERE clauseID IN (SELECT clauseID from used_cl_ids );
        """

        self.c.execute(q)
        print("Filtered into used_clauses_red T: %-3.2f s" % (time.time() - t))

    def drop_used_clauses_red(self):
        q = """
        drop TABLE if exists used_clauses_red;
        """
        self.c.execute(q)

    def filter_tables_of_ids(self):

        queries = """
        drop index if exists idxclid34;
        drop index if exists idxclid32;
        """
        for q in queries.split("\n"):
            self.c.execute(q)

        self.print_idxs()

        tables = ["clause_stats", "reduceDB", "sum_cl_use", "used_clauses_anc",
                  "used_clauses", "restart_dat_for_cl", "cl_last_in_solver"]
        q = """
        DELETE FROM {table} WHERE clauseID NOT IN
        (SELECT clauseID from used_cl_ids );"""

        for table in tables:
            t = time.time()
            self.c.execute(q.format(table=table))
            print("Filtered table '%s' T: %-3.2f s" % (table, time.time() - t))

    def print_sum_cl_use_distrib(self):
        q = """
        select c, num_used from (
        select count(*) as c, num_used
        from sum_cl_use
        group by num_used) order by num_used
        """
        self.c.execute(q)
        rows = self.c.fetchall()
        print("Distribution of clause uses:")
        total = 0
        zero_use = 0
        for i in range(len(rows)):
            cnt = int(rows[i][0])
            numuse = int(rows[i][1])
            if numuse == 0:
                zero_use = cnt
            total += cnt

        i = 0
        while i < len(rows):
            cnt = int(rows[i][0])
            numuse = int(rows[i][1])

            this_cnt_tot = 0
            this_numuse_tot = 0
            for x in range(100000):
                if i+x >= len(rows):
                    i+=x
                    break

                this_cnt = int(rows[i+x][0])
                this_numuse = int(rows[i+x][1])
                this_cnt_tot += this_cnt
                this_numuse_tot += this_numuse
                if this_cnt_tot > 300:
                    i+=x
                    i+=1
                    break
            print("  ->  {cnt:-8d} of sum_cl_use: {numuse:-7d}-{this_numuse:-7d}  --  {percent:-3.5f} ratio".format(
                    cnt=this_cnt_tot, numuse=numuse, this_numuse=this_numuse, percent=(this_cnt_tot/total)))


        print("Total: %d of which zero_use: %d" % (total, zero_use))
        # grain-53-80-0s0-seed-125-4-init-35.cnf.out actually has only 0.0098 use (i.e. 0.98%)
        if zero_use == 0 or zero_use/total < 0.009:
            print("ERROR: Zero use is very low, this is almost surely a bug!")
            exit(-1)

    def check_db_sanity(self):
        print("Checking tables in DB...")
        q = """
        SELECT name FROM sqlite_master WHERE type == 'table'
        """
        found_sum_cl_use = False
        self.c.execute(q)
        rows = self.c.fetchall()
        for row in rows:
            if row[0] == "sum_cl_use":
                found_sum_cl_use = True

            print("-> We have table: ", row[0])
            if row[0] == "used_later_short" or row[0] == "used_later_long":
                print("ERROR: 'gen_pandas.py' has been already ran on this DB")
                print("       this will be a mess. We cannot run. ")
                print("       Exiting.")
                exit(-1)

        if not found_sum_cl_use:
            print("ERROR: Did not find sum_cl_use table. You probably didn't run")
            print("       the 'clean_update_data.py' on this database")
            print("       Exiting.")
            exit(-1)

        q = """
        SELECT count() FROM sum_cl_use where num_used = 0
        """
        self.c.execute(q)
        rows = self.c.fetchall()
        assert len(rows) == 1
        num = int(rows[0][0])
        print("Unused clauses in sum_cl_use: ", num)
        if num == 0:
            print("ERROR: You most likely didn't run 'clean_data.py' on this database")
            print("       Exiting.")
            exit(-1)

        print("Tables seem OK")

    def insert_into_only_keep_rdb(self, min_used_later, limit, tier, table):
        limit = int(limit)
        t = time.time()
        q = """
        insert into only_keep_rdb (id)
        select
        rdb0.rowid

        FROM
        reduceDB as rdb0,
        {table}_{tier}

        WHERE
        {table}_{tier}.clauseID=rdb0.clauseID
        and {table}_{tier}.rdb0conflicts=rdb0.conflicts
        and {table}_{tier}.used_later >= {min_used_later}
        order by random()
        limit {limit}""".format(min_used_later=min_used_later, limit=limit,
                                tier=tier, table=table)
        self.c.execute(q)

        ret = self.c.execute("""select count() from only_keep_rdb""")
        rows = self.c.fetchall()
        rdb_rows = rows[0][0]

        print("Insert only_keep_rdb where %s_%s >= %d T: %-3.2f s. Now size: %s" %
              (table, tier, min_used_later, time.time() - t, rdb_rows))

    def delete_too_many_rdb_rows(self):
        t = time.time()
        val = int(options.limit)
        ret = self.c.execute("select count() from reduceDB")
        rows = self.c.fetchall()
        rdb_rows = rows[0][0]
        print("Have %d lines of RDB, let's do non-fair selection" % (rdb_rows))

        q = """
        drop table if exists only_keep_rdb;
        """
        self.c.execute(q)

        t = time.time()
        q = """create table only_keep_rdb (
            id bigint(20) not null
        );"""
        self.c.execute(q)
        print("Created only_keep_rdb T: %-3.2f s" % (time.time() - t))

        mygoal = options.goal_rdb/13
        table="used_later" # we could iterate with "used_later_anc", but not doing that
        for tier in ["short", "long", "forever"]:
            mygoal*=3 # this gives 8%, 23%, 70% distribution (total 100% if not rounded)
            if not options.fair:
                self.insert_into_only_keep_rdb(100000, mygoal/20, tier=tier, table=table)
                self.insert_into_only_keep_rdb(10000, mygoal/20, tier=tier, table=table)
                self.insert_into_only_keep_rdb(1000, mygoal/20, tier=tier, table=table)
                self.insert_into_only_keep_rdb(100, mygoal/10, tier=tier, table=table)
                self.insert_into_only_keep_rdb(20, mygoal/5, tier=tier, table=table)
                self.insert_into_only_keep_rdb(10, mygoal/5, tier=tier, table=table)
                self.insert_into_only_keep_rdb(5, mygoal/3, tier=tier, table=table)
                self.insert_into_only_keep_rdb(1, mygoal/2, tier=tier, table=table)
            self.insert_into_only_keep_rdb(0, mygoal, tier=tier, table=table)

        t = time.time()
        ret = self.c.execute("select count() from only_keep_rdb")
        rows = self.c.fetchall()
        rdb_rows = rows[0][0]
        print("We now have %d lines only_keep_rdb" % (rdb_rows))


        t = time.time()
        q = """
        drop index if exists `idxclid6-4`; -- the other index on reduceDB
        create index `idx_bbb` on `only_keep_rdb` (`id`);
        """
        for l in q.split('\n'):
            self.c.execute(l)
        print("only_keep_rdb indexes added T: %-3.2f s" % (time.time() - t))

        q = """
        delete from reduceDB
        where reduceDB.rowid not in (select id from only_keep_rdb)
        """
        self.c.execute(q)
        print("Delete from reduceDB T: %-3.2f s" % (time.time() - t))

        t = time.time()
        val = int(options.limit)
        ret = self.c.execute("select count() from reduceDB")
        rows = self.c.fetchall()
        rdb_rows = rows[0][0]
        print("Finally have %d lines of RDB" % (rdb_rows))


    def del_table_and_vacuum(self):
        helper.drop_idxs(self.c)

        t = time.time()
        queries = """
        DROP TABLE IF EXISTS `used_later`;
        DROP TABLE IF EXISTS `only_keep_rdb`;
        DROP TABLE IF EXISTS `used_cl_ids`;
        """
        for q in queries.split("\n"):
            self.c.execute(q)
        print("Deleted tables T: %-3.2f s" % (time.time() - t))

        q = """
        vacuum;
        """

        t = time.time()
        lev = self.conn.isolation_level
        self.conn.isolation_level = None
        self.c.execute(q)
        self.conn.isolation_level = lev
        print("Vacuumed database T: %-3.2f s" % (time.time() - t))


if __name__ == "__main__":
    usage = "usage: %prog [options] sqlitedb"
    parser = optparse.OptionParser(usage=usage)

    parser.add_option("--limit", default=20000, type=int,
                      dest="limit", help="Number of clauses to limit ourselves to")
    parser.add_option("--goalrdb", default=200000, type=int,
                      dest="goal_rdb", help="Number of RDB neeeded")
    parser.add_option("--goalvardata", default=50000, type=int,
                      dest="goal_vardata", help="Number of varData points neeeded")
    parser.add_option("--verbose", "-v", action="store_true", default=False,
                      dest="verbose", help="Print more output")
    parser.add_option("--noidx", action="store_true", default=False,
                      dest="noidx", help="Don't recreate indexes")
    parser.add_option("--fair", "-f", action="store_true", default=False,
                      dest="fair", help="Fair sampling. NOT DEFAULT.")

    # lengths of short/long
    parser.add_option("--short", default=10*1000, type=int,
                      dest="short", help="Short duration. Default: %default")
    parser.add_option("--long", default=30*1000, type=int,
                      dest="long", help="Long duration. Default: %default")
    parser.add_option("--forever", default=120*1000, type=int,
                      dest="forever", help="Forever duration. Default: %default")

    (options, args) = parser.parse_args()

    if len(args) < 1:
        print("ERROR: You must give the sqlite file!")
        exit(-1)

    if not options.fair:
        print("NOTE: Sampling will NOT be fair.")
        print("      This is because otherwise, DB will be huge")
        print("      and we need lots of positive datapoints")
        print("      most of which will be from clauses that are more used")

    with QueryDatRem(args[0]) as q:
        q.check_db_sanity()
        helper.dangerous(q.c)
        helper.drop_idxs(q.c)
        q.create_indexes1()
        q.remove_too_many_vardata()

    # this is the SLOW way of doing it -- without pre-sampling it
    if False:
        print("This is good for verifying that the fast ones are close")
        # slower percentiles
        t = time.time()
        with helper.QueryFill(args[0]) as q:
            helper.dangerous(q.c)
            q.delete_and_create_used_laters()
            q.create_indexes(verbose=options.verbose)
            for tier in ["short", "long", "forever"]:
                q.fill_used_later_X(tier, getattr(options, tier))
        with QueryDatRem(args[0]) as q:
            helper.dangerous(q.c)
            q.create_percentiles_table()
            for tier in ["short", "long", "forever"]:
                q.get_all_percentile_X(tier)
            q.print_percentiles()
        with helper.QueryFill(args[0]) as q:
            q.delete_and_create_used_laters()
        print("SLOWER percentiles:", time.time()-t)

    # Percentile generation
    t = time.time()

    # first, we create "used_clauses_red" that contains a reduced
    # list of clauseIDs we want to do this over. Here we sample FAIRLY!!!!
    with QueryDatRem(args[0]) as q:
        helper.dangerous(q.c)
        q.recreate_used_ID_table()
        q.fill_used_cl_ids_table(True, limit=4*options.limit) # notice the FAIR sampling!
        q.drop_used_clauses_red()
        q.create_used_clauses_red()

    # now we generate the used_later data
    with helper.QueryFill(args[0]) as q:
        helper.dangerous(q.c)
        q.delete_and_create_used_laters()
        q.create_indexes(verbose=options.verbose, used_clauses="used_clauses_red")
        for table in ["used_later", "used_later_anc"]:
            for tier in ["short", "long", "forever"]:
                q.fill_used_later_X(tier, duration=getattr(options, tier),
                                    used_clauses="used_clauses_red",
                                    table=table)

    # now we calculate the distributions and save them
    with QueryDatRem(args[0]) as q:
        helper.dangerous(q.c)
        q.create_percentiles_table()
        for tier in ["short", "long", "forever"]:
                q.get_all_percentile_X(tier)
        q.print_percentiles()
        q.drop_used_clauses_red()
    with helper.QueryFill(args[0]) as q:
        helper.drop_idxs(q.c)
        q.delete_and_create_used_laters()
    print("FASTER percentiles:", time.time()-t)

    # Filtering for clauseIDs in tables:
    # ["clause_stats", "reduceDB", "sum_cl_use",
    #      "used_clauses", "restart_dat_for_cl", "cl_last_in_solver"]
    # here we sample NON-FAIRLY (options.fair) by default!
    with QueryDatRem(args[0]) as q:
        helper.dangerous(q.c)
        q.recreate_used_ID_table()
        q.fill_used_cl_ids_table(options.fair, limit=options.limit)
        q.filter_tables_of_ids()
        q.print_sum_cl_use_distrib()
        print("-------------")

    # RDB filtering
    with helper.QueryFill(args[0]) as q:
        helper.dangerous(q.c)
        helper.drop_idxs(q.c)
        q.delete_and_create_used_laters()
        q.create_indexes(verbose=options.verbose)

        # this is is needed for RDB row deletion below (since it's not fair)
        table = "used_later" # we could also do used_later_anc (for RDB)
        for tier in ["short", "long", "forever"]:
            q.fill_used_later_X(tier=tier, table=table, duration=getattr(options, tier))

    with QueryDatRem(args[0]) as q:
        print("-------------")
        q.delete_too_many_rdb_rows() # NOTE: NOT FAIR!!!

        helper.drop_idxs(q.c)
        q.del_table_and_vacuum()



#!/bin/python3

import sys
import os

assert len(sys.argv) >= 2


tiers = {"short":[], "long":[], "forever":[]}

for mydir in sys.argv[1:]:
    if not os.path.isdir(mydir):
        print("WARN: %s is not a directory, skipping" % mydir)
        continue
    print(mydir)
    print("========")
    for tier in ["short", "long", "forever"]:
        for table in ["used_later", "used_later_anc"]:
            fname = "%s/out-%s-%s" %(mydir, table, tier)
            if not os.path.isfile(fname):
                print("ERROR: the file '%s' does not exist!" % fname)
                exit(-1)

            test = False
            all_dump_no = False
            with open(fname, "r")  as f:
                out = "Tier %-9s " % tier
                for line in f:
                    line = line.strip()
                    if " test data" in line:
                        test = True

                    if not test:
                        continue

                    if "ALL dump_no" in line:
                        all_dump_no = True

                    if not all_dump_no:
                        continue

                    if "train+test" in line:
                        break

                    if "train data" in line:
                        break

                    assert test
                    if "Mean squared error is" in line:
                        mydat = line.split(" ")
                        val = float(mydat[5])
                        tiers[tier].append(["%s-%s" % (mydir, table), val])
                        out += line
                        print(out)
                        break

for tier in ["short", "long", "forever"]:
    print("TIER: ", tier)
    top =  sorted(tiers[tier], key=lambda x: x[1], reverse=False)
    for t in top:
        print(t)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2009-2020 Authors of CryptoMiniSat, see AUTHORS file
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA

from __future__ import with_statement  # Required in 2.5
from __future__ import print_function
import optparse
import gzip
import re
import fnmatch
from xor_to_cnf_class import *
from debuglib import *
import subprocess
import os
import stat
import time
import resource
from functools import partial


def unique_file(fname_begin, fname_end=".cnf"):
        counter = 1
        while 1:
            fname = "out/" + fname_begin + '_' + str(counter) + fname_end
            try:
                fd = os.open(
                    fname, os.O_CREAT | os.O_EXCL, stat.S_IREAD | stat.S_IWRITE)
                os.fdopen(fd).close()
                return fname
            except OSError:
                pass

            counter += 1
            if counter > 300:
                print("Cannot create unique_file, last try was: %s" % fname)
                exit(-1)


def setlimits(maxtime):
    # sys.stdout.write("Setting resource limit in child (pid %d): %d s\n" %
    # (os.getpid(), maxtime))
    resource.setrlimit(resource.RLIMIT_CPU, (maxtime, maxtime))


class solution_parser:
    def __init__(self, options):
        self.options = options

    @staticmethod
    def test_found_solution(solution, fname, debugLibPart=None):
        if debugLibPart is None:
            print("Verifying solution for CNF file %s" % fname)
        else:
            print("Verifying solution for CNF file %s, part %d" %
                  (fname, debugLibPart))

        if fnmatch.fnmatch(fname, '*.gz'):
            f = gzip.open(fname, "r")
        else:
            f = open(fname, "r")
        clauses = 0
        thisDebugLibPart = 0

        for line in f:
            line = line.rstrip()

            # skip empty lines
            if len(line) == 0:
                continue

            # count debug lib parts
            if line[0] == 'c' and "Solver::solve" in line:
                thisDebugLibPart += 1

            # if we are over debugLibPart, exit
            if debugLibPart is not None and thisDebugLibPart >= debugLibPart:
                f.close()
                return

            # check solution against clause
            try:
                if line[0] != 'c' and line[0] != 'p':
                    if line[0] != 'x':
                        solution_parser._check_regular_clause(line, solution)
                    else:
                        assert line[0] == 'x', "Line must start with p, c, v or x"
                        solution_parser._check_xor_clause(line, solution)

                clauses += 1
            except:
                if debugLibPart is not None:
                    print("--> Error in part: %s. We are reading up to and including part: %s"
                          % (thisDebugLibPart, debugLibPart-1))
                raise

        f.close()
        print("Verified %d original xor&regular clauses" % clauses)

    def sampling_vars_solution_check(self, fname, sampling_vars, solution):
        assert len(sampling_vars) > 0
        a = XorToCNF()
        tmpfname = unique_file("tmp_for_xor_to_cnf_convert")
        a.convert(fname, tmpfname, len(sampling_vars))

        with open(tmpfname, "a") as f:
            # NOTE: the "p cnf..." header will be wrong
            for i in sampling_vars:
                if i not in solution:
                    print("ERROR: solution does not contain sampling var %d" % i)
                    print("Sampling vars were: %s" % sampling_vars)
                    exit(-1)

                if solution[i]:
                    f.write("%d 0\n" % i)
                else:
                    f.write("-%d 0\n" % i)

        print("-> added partial solution to temporary CNF file %s" % tmpfname)

        # execute with the other solver
        toexec = "../../build/utils/lingeling-ala/lingeling %s" % tmpfname
        print("sampling check -- solving with other solver: %s" % toexec)
        curr_time = time.time()
        try:
            p = subprocess.Popen(toexec.rsplit(),
                                 stdout=subprocess.PIPE,
                                 preexec_fn=partial(setlimits, self.options.maxtime),
                                 universal_newlines=True)
        except OSError:
            print("ERROR: Lingeing didn't run... did you install it? It should be in the path as `lingeling`")
            raise

        consoleOutput2 = p.communicate()[0]
        os.unlink(tmpfname)

        # if other solver was out of time, then we can't say anything
        diff_time = time.time() - curr_time
        if diff_time > self.options.maxtime - self.options.maxtimediff:
            print("Other solver: too much time to solve, aborted!")
            return None

        # extract output from the other solver
        print("Checking other solver output...")
        otherSolverUNSAT, _, _ = self.parse_solution_from_output(
            consoleOutput2.split("\n"))

        # check if the other solver finds a solution with the sampling vars
        # set as per partial solution returned
        if otherSolverUNSAT is True:
            print("ERROR; The other solver did NOT find a solution with the partial solution given")
            exit(-1)
            return False

        print("OK, other solver found a solution using the partial solution")

        return True

    def check_unsat(self, fname):
        a = XorToCNF()
        tmpfname = unique_file("tmp_for_xor_to_cnf_convert")
        a.convert(fname, tmpfname)

        # execute with the other solver
        toexec = "../../build/utils/lingeling-ala/lingeling %s" % tmpfname
        print("UNSAT check -- solving with other solver: %s" % toexec)
        curr_time = time.time()
        try:
            p = subprocess.Popen(toexec.rsplit(),
                                 stdout=subprocess.PIPE,
                                 preexec_fn=partial(setlimits, self.options.maxtime),
                                 universal_newlines=True)
        except OSError:
            print("ERROR: Lingeling didn't run... did you install it? It should be in the path as `lingeling`")
            raise

        consoleOutput2 = p.communicate()[0]
        os.unlink(tmpfname)

        # if other solver was out of time, then we can't say anything
        diff_time = time.time() - curr_time
        if diff_time > self.options.maxtime - self.options.maxtimediff:
            print("Other solver: too much time to solve, aborted!")
            return None

        # extract output from the other solver
        print("Checking other solver output...")
        otherSolverUNSAT, otherSolverSolution, _ = self.parse_solution_from_output(
            consoleOutput2.split("\n"))

        # check if the other solver agrees with us
        return otherSolverUNSAT

    def check_debug_lib(self, fname, must_check_unsat=True):
        largestPart = self._find_largest_debuglib_part(fname)
        for debugLibPart in range(1, largestPart + 1):
            fname_debug = "%s-debugLibPart%d.output" % (fname, debugLibPart)
            print("Checking debug lib part %s -- %s " % (debugLibPart, fname_debug))

            if (os.path.isfile(fname_debug) is False):
                print("Error: Filename to be read '%s' is not a file!" % fname_debug)
                exit(-1)

            # take file into mem
            f = open(fname_debug, "r")
            text = f.read()
            output_lines = text.splitlines()
            f.close()

            unsat, solution, conflict = self.parse_solution_from_output(output_lines)
            assumps = self._get_assumps(fname, debugLibPart)
            if unsat is False:
                print("debugLib is SAT")
                self._check_assumps_inside_solution(assumps, solution)
                self.test_found_solution(solution, fname, debugLibPart)
            else:
                print("debugLib is UNSAT")
                if must_check_unsat:
                    assert conflict is not None, "debugLibPart must create a conflict in case of UNSAT"
                    self._check_assumps_inside_conflict(assumps, conflict)
                    tmpfname = unique_file("tmp_for_extract_libpart")
                    self._extract_lib_part(fname, debugLibPart, assumps, tmpfname)

                    # check with other solver
                    ret = self.check_unsat(tmpfname)
                    if ret is None:
                        print("Cannot check, other solver took too much time")
                    elif ret is True:
                        print("UNSAT verified by other solver")
                    else:
                        print("Grave bug: SAT-> UNSAT : Other solver found solution!!")
                        exit(-1)
                    os.unlink(tmpfname)

        self.remove_debuglib_files(fname)

    def remove_debuglib_files(self, fname):
        #removing debuglib files
        largestPart = self._find_largest_debuglib_part(fname)
        for debugLibPart in range(1, largestPart + 1):
            fname_debug = "%s-debugLibPart%d.output" % (fname, debugLibPart)
            os.unlink(fname_debug)

    @staticmethod
    def parse_solution_from_output(output_lines, ignoreNoSolution=False):
        if len(output_lines) == 0:
            print("Error! SAT solver output is empty!")
            print("output lines: %s" % output_lines)
            exit(-1)

        # solution will be put here
        satunsatfound = False
        vlinefound = False
        solution = {}
        conflict = None

        # parse in solution
        for line in output_lines:
            # skip comment
            if re.match('^conflict ', line):
                line = line.strip().split()[1:]
                conflict = [int(elem) for elem in line]
                continue

            if (re.match('^c ', line)):
                continue

            # solution
            if (re.match('^s ', line)):
                if (satunsatfound):
                    print("ERROR: solution twice in solver output!")
                    exit(400)

                if 'UNSAT' in line:
                    unsat = True
                    satunsatfound = True
                    continue

                if 'SAT' in line:
                    unsat = False
                    satunsatfound = True
                    continue

                print("ERROR: line starts with 's' but no SAT/UNSAT on line")
                exit(400)

            # parse in solution
            if (re.match('^v ', line)):
                vlinefound = True
                myvars = line.split(' ')
                for var in myvars:
                    var = var.strip()
                    if var == "" or var == 'v':
                        continue
                    if (int(var) == 0):
                        break
                    intvar = int(var)
                    solution[abs(intvar)] = (intvar >= 0)
                continue

            if (line.strip() == ""):
                continue

            print("Error! SAT solver output contains a line that is neither 'v' nor 'c' nor 's'!")
            print("Line is:", line.strip())
            exit(-1)

        # print("Parsed values:", solution)

        if (ignoreNoSolution is False and
                (satunsatfound is False or (
                    unsat is False and vlinefound is False))):
            print("Error: Cannot find line starting with 's' or 'v' in output!")
            print(output_lines)
            print("Error code 500")
            exit(-1)

        if (ignoreNoSolution is True and
                (satunsatfound is False or (
                    unsat is False and vlinefound is False))):
            print("Probably timeout, since no solution  printed. Could, of course, be segfault/assert fault, etc.")
            print("Making it look like an UNSAT, so no checks!")
            return (True, [])

        if (satunsatfound is False):
            print("Error: Cannot find if SAT or UNSAT. Maybe didn't finish running?")
            print(output_lines)
            print("Error code 500")
            exit(-1)

        if (unsat is False and vlinefound is False):
            print("Error: Solution is SAT, but no 'v' line")
            print (output_lines)
            print("Error code 500")
            exit(-1)

        return unsat, solution, conflict

    def _extract_lib_part(self, fname, debug_num, assumps, tofile):
        fromf = open(fname, "r")
        thisDebugLibPart = 0
        maxvar = 0
        numcls = 0
        for line in fromf:
            line = line.strip()

            # ignore empty strings and headers
            if not line or line[0] == "p":
                continue

            # process (potentially special) comments
            if line[0] == "c":
                if "Solver::solve" in line:
                    thisDebugLibPart += 1

                continue

            # break out if we reached the debug lib part
            if thisDebugLibPart >= debug_num:
                break

            # count clauses and get max var number
            numcls += 1
            maxvar = max(maxvar, get_max_var_from_clause(line))

        fromf.close()

        # now we can create the new CNF file
        fromf = open(fname, "r")
        tof = open(tofile, "w")
        tof.write("p cnf %d %d\n" % (maxvar, numcls + len(assumps)))

        thisDebugLibPart = 0
        for line in fromf:
            line = line.strip()
            # skip empty lines and headers
            if not line or line[0] == "p":
                continue

            # parse up special header
            if line[0] == "c":
                if "Solver::solve" in line:
                    thisDebugLibPart += 1

                continue

            # break out if we reached the debug lib part
            if thisDebugLibPart >= debug_num:
                break

            tof.write(line + '\n')

        # add assumptions
        for lit in assumps:
            tof.write("%d 0\n" % lit)

        fromf.close()
        tof.close()

    def _get_assumps(self, fname, debugLibPart):
        f = open(fname, "r")

        thispart = 0
        solveline = None
        for line in f:
            if "Solver::solve" in line:
                thispart += 1
                if thispart == debugLibPart:
                    solveline = line
                    break
        f.close()

        assert solveline is not None
        ret = re.match("c.*Solver::solve\((.*)\)", solveline)
        assert ret is not None
        assumps = ret.group(1).strip().split()
        assumps = [int(x) for x in assumps]

        print("Assumptions: ", assumps)
        return assumps

    def _check_assumps_inside_conflict(self, assumps, conflict):
        for lit in conflict:
            if -1 * lit not in assumps:
                print("ERROR: Final conflict contains %s but assumps is %s" %(conflict, assumps))
                print("ERROR: lit ", lit, " is in conflict but its inverse is not is assumps!")
                exit(-100)

        print("OK, final conflict only contains elements from assumptions")

    def _check_assumps_inside_solution(self, assumps, solution):
        for lit in assumps:
            var = abs(lit)
            val = lit > 0
            if var in solution:
                if solution[var] != val:
                    print("Solution pinted has literal %s but assumptions contained the inverse: '%s'" % (-1 * lit, assumps))
                    exit(-100)

        print("OK, all assumptions inside solution")

    def _find_largest_debuglib_part(self, fname):
        largestPart = 0
        dirList2 = os.listdir("out/.")
        fname_no_out = fname.replace("out/", "")
        for fname_debug in dirList2:
            if fnmatch.fnmatch(fname_debug, "%s-debugLibPart*.output" % fname_no_out):
                largestPart += 1

        return largestPart

    @staticmethod
    def max_vars_in_file(fname):
        maxvar = 0
        with open(fname, "r") as f:
            for line in f:
                line = line.strip()

                # ignore comments
                if not line or line[0] == "c" or line[0] == 'p':
                    continue

                # calculate max variable
                maxvar = max(maxvar, get_max_var_from_clause(line))

        return maxvar

    @staticmethod
    def _check_regular_clause(line, solution):
        lits = line.split()
        for lit in lits:
            numlit = int(lit)
            if numlit == 0:
                break

            if abs(numlit) not in solution:
                continue

            if solution[abs(numlit)] ^ (numlit < 0):
                return True

        # print not set vars
        for lit in lits:
            numlit = int(lit)
            if numlit == 0:
                break

            if abs(numlit) not in solution:
                print("var %d in XOR clause not set" % abs(numlit))

        #print("Every other var set to FALSE")
        #print("Orig clause in DIMACS: ", lits)
        print("Error: clause '%s' not satisfied." % line.strip())
        raise NameError("Error: clause '%s' not satisfied." % line)
        return False

    @staticmethod
    def _check_xor_clause(line, solution):
        line = line.lstrip('x')
        lits = line.split()
        final = False
        for lit in lits:
            numlit = int(lit)
            if numlit != 0:
                if abs(numlit) not in solution:
                    raise NameError("Error: var %d not solved, but referred to in a xor-clause of the CNF" % abs(numlit))
                final ^= solution[abs(numlit)]
                final ^= numlit < 0
        if final is False:
            print("Error: xor-clause '%s' not satisfied." % line.strip())
            raise NameError("Error: xor-clause '%s' not satisfied." % line)

        return final


def parse_arguments():
    class PlainHelpFormatter(optparse.IndentedHelpFormatter):

        def format_description(self, description):
            if description:
                return description + "\n"
            else:
                return ""

    usage = """usage: %prog solution cnf

For example:
%prog my_solution_file.out my_problem.cnf.gz"""
    parser = optparse.OptionParser(usage=usage, formatter=PlainHelpFormatter())
    parser.add_option("--verbose", "-v", action="store_true",
                      default=False, dest="verbose", help="Be more verbose")
    parser.add_option("--tout", "-t", dest="maxtime", type=int, default=100,
                      help="Max time to run. Default: %default")
    parser.add_option("--textra", dest="maxtimediff", type=int, default=10,
                      help="Extra time on top of timeout for processing."
                      " Default: %default")
    # parse options
    options, args = parser.parse_args()
    return options, args

if __name__ == "__main__":
    options, args = parse_arguments()
    print("Options are:", options)
    print("args are:", args)
    if len(args) != 2:
        print("ERROR: You must give exactly two parameters, "
              "one SOLUTION and one CNF")
        print("You gave {n} parameters".format(**{"n": len(args)}))
        exit(-1)

    sol_file = args[0]
    cnf_file = args[1]
    print("Verifying CNF file '{cnf}' against solution in file '{sol}'".format(
        **{"cnf": cnf_file, "sol": sol_file}))

    print("Checking debug libs...")
    sol_parser = solution_parser(options)
    sol_parser.check_debug_lib(cnf_file)

    print("Checking console output...")
    sol = {}
    with open(sol_file) as f:
        dat = f.read()

    dat = dat.split("\n")
    unsat, solution, _ = sol_parser.parse_solution_from_output(dat)
    if not unsat:
        sol_parser.test_found_solution(solution, cnf_file)
        exit(0)

    # check with other solver
    ret = sol_parser.check_unsat(cnf_file)
    if ret is None:
        print("Other solver time-outed, cannot check")
    elif ret is True:
        print("UNSAT verified by other solver")
    else:
        print("Grave bug: SAT-> UNSAT : Other solver found solution!!")
        exit(-1)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2009-2020 Authors of CryptoMiniSat, see AUTHORS file
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import operator
import re
import pandas as pd
import pickle
import sklearn
import sklearn.svm
import sklearn.tree
import sklearn.cluster
from sklearn.preprocessing import StandardScaler
import argparse
import sys
import numpy as np
import sklearn.metrics
import time
import itertools
import math
import matplotlib.pyplot as plt
import sklearn.ensemble
import os
import helper
import decimal
try:
    from termcolor import cprint
except ImportError:
    termcolor_avail = False
else:
    termcolor_avail = True

ver = sklearn.__version__.split(".")
if int(ver[1]) < 20:
    from sklearn.cross_validation import train_test_split
else:
    from sklearn.model_selection import train_test_split


CLUST_TYPE_FILES = 0
CLUST_TYPE_USEFULNESS = 1

def get_cluster_name(clust_type):
    name = None
    if clust_type == CLUST_TYPE_FILES:
        name = "clust_f"
    elif clust_type == CLUST_TYPE_USEFULNESS:
        name = "clust_u"
    else:
        assert False

    return name

def get_clustering_label(df, feats_used: list, scaler, clustering):
    this_feats = list(df)
    for f in feats_used:
        if f not in this_feats:
            print("ERROR: Feature '%s' is in the training file but not in the current one!" % f)
            exit(-1)

    df_clust = df[feats_used].astype(float).copy()
    df_clust = scaler.transform(df_clust)
    return clustering.predict(df_clust)


def dump_dataframe(df, fname: str):
    with open(fname, "wb") as f:
        pickle.dump(df, f)
    print("Dumped to file %s" % fname)


class Clustering:
    def __init__(self, df, clust_type):
        self.df = df
        self.clust_type = clust_type

    def reformat_feats(self, feats):
        feats_clean = []
        for feat in feats:
            assert "szfeat_cur.conflicts" not in feat
            c = str(feat)
            c = c.replace(".irred_", ".irred_cl_distrib.")
            c = c.replace(".red_", ".red_cl_distrib.")
            c = c.replace("szfeat_cur.", "{val}.")

            feats_clean.append(c)
        assert len(feats_clean) == len(feats)

        return feats_clean

    def create_code_for_cluster_centers(self, clust, scaler, sz_feats):
        sz_feats_clean = self.reformat_feats(sz_feats)
        f = open("{basedir}/clustering_imp.cpp".format(basedir=options.basedir), 'w')

        helper.write_mit_header(f)
        f.write("""
#include "satzilla_features.h"
#include "clustering.h"
#include <cmath>

using namespace CMSat;

ClusteringImp::ClusteringImp() {{
    set_up_centers();
}}

ClusteringImp::~ClusteringImp() {{
}}

""".format(clusters=options.clusters))

        f.write("void ClusteringImp::set_up_centers() {\n")
        f.write("\n        centers.resize(%d);\n" % options.clusters)
        for i in range(options.clusters):
            f.write("\n        // Doing cluster center %d\n" % i)
            f.write("\n        centers[%d].resize(%d);\n" % (i, len(sz_feats_clean)) )
            f.write("\n        used_clusters.push_back(%d);\n" % i)
            for i2 in range(len(sz_feats_clean)):
                feat = sz_feats_clean[i2]
                center = clust.cluster_centers_[i][i2]
                f.write("        centers[{num}][{feat}] = {center}L;\n".format(
                    num=i, feat=i2, center=center))

        f.write("    }\n")

        f.write("""
double sq(double x) {
    return x*x;
}

double ClusteringImp::norm_dist(const SatZillaFeatures& a, const std::vector<double>& center) const {
    double dist = 0;
    double tmp;
""")
        for feat, i in zip(sz_feats_clean, range(10000)):
            printed_scale = "{scale:3.10f}".format(scale=scaler.scale_[i])
            if decimal.Decimal(printed_scale) == decimal.Decimal("0.0") :
                    f.write("         // feature {feat} would have caused a division by zero, avoiding\n".\
                        format(feat=feat.format(val="a")))
                    continue

            f.write("        tmp = ((double){feat}-{mean:3.9f})/{printed_scale};\n".format(
                feat=feat.format(val="a"),
                 mean=scaler.mean_[i],
                 printed_scale=printed_scale
                 ))

            f.write("    dist+=sq(tmp-center[{feat}]);\n\n".format(
                feat=i
                ))

        f.write("""
    return dist;
}\n""")

        f.write("""
int ClusteringImp::which_is_closest(const SatZillaFeatures& p) const {
    double closest_dist = numeric_limits<double>::max();
    int closest = -1;
    for (int i: used_clusters) {
        double dist = norm_dist(p, centers[i]);
        if (dist < closest_dist) {
            closest_dist = dist;
            closest = i;
        }
    }
    return closest;
    }\n""")

    def select_features_files(self):
        features = list(self.df)
        for f in list(features):
            if any(ext in f for ext in ["var", "vcg", "pnr", "min", "max", "std"]):
                features.remove(f)

        # features from dataframe
        feats_used = []
        for feat in features:
            if "szfeat_cur" in feat and "szfeat_cur.conflicts" not in feat:
                feats_used.append(feat)

        return feats_used

    def select_features_usefulness(self):
        features = list(self.df)
        feats_used = helper.get_features(options.best_features_fname)

        return feats_used

    def cluster(self):
        if self.clust_type == CLUST_TYPE_FILES:
            self.feats_used = self.select_features_files()
        elif self.clust_type == CLUST_TYPE_USEFULNESS:
            self.feats_used = self.select_features_usefulness()
        else:
            assert False

        print("Features used: ", self.feats_used)
        print("Number of features used: ", len(self.feats_used))

        if options.check_row_data:
            helper.check_too_large_or_nan_values(self.df, self.feats_used)
        else:
            helper.check_too_large_or_nan_values(self.df.sample(100), self.feats_used)

        # fit to slice that only includes CNF features
        df_clust = self.df[self.feats_used].astype(float).copy()
        if options.scale:
            self.scaler = StandardScaler()
            self.scaler.fit(df_clust)
            if options.verbose:
                print("Scaler:")
                print(" -- ", self.scaler.mean_)
                print(" -- ", self.scaler.scale_)

            if options.verbose:
                df_clust_back = df_clust.copy()
            df_clust = self.scaler.transform(df_clust)
        else:
            class ScalerNone:
                def __init__(self):
                    self.mean_ = [0.0 for n in range(df_clust.shape[1])]
                    self.scale_ = [1.0 for n in range(df_clust.shape[1])]
            self.scaler = ScalerNone()

        # test scaler's code generation
        if options.scale and options.verbose:
            # we rely on this later in code generation
            # for scaler.mean_
            # for scaler.scale_
            # for cluster.cluster_centers_
            for i in range(df_clust_back.shape[1]):
                assert df_clust_back.columns[i] == self.feats_used[i]

            # checking that scaler works as expected
            for feat in range(df_clust_back.shape[1]):
                df_clust_back[df_clust_back.columns[feat]
                              ] -= self.scaler.mean_[feat]
                df_clust_back[df_clust_back.columns[feat]
                              ] /= self.scaler.scale_[feat]

            print(df_clust_back.head()-df_clust.head())
            print(df_clust_back.head())
            print(df_clust.head())

        self.clust = sklearn.cluster.KMeans(n_clusters=options.clusters, random_state=prng)
        self.clust.fit(df_clust)
        self.df["clust"] = self.clust.labels_

        # print information about the clusters
        if options.verbose:
            print(self.feats_used)
            print(self.clust.labels_)
            print(self.clust.cluster_centers_)
            print(self.clust.get_params())

        # code for cluster centers
        self.create_code_for_cluster_centers(
            self.clust, self.scaler, self.feats_used)

        return self.feats_used, self.scaler, self.clust


if __name__ == "__main__":
    usage = "usage: %(prog)s [options] file.pandas"
    parser = argparse.ArgumentParser(usage=usage)

    parser.add_argument("--seed", default=None, type=int,
                        dest="seed", help="Seed of PRNG")
    parser.add_argument("fnames", nargs='+', metavar='FILES',
                        help="Files to parse, first file is the one to base data off of")
    parser.add_argument("--verbose", "-v", action="store_true", default=False,
                        dest="verbose", help="Print more output")
    parser.add_argument("--printfeat", action="store_true", default=False,
                        dest="print_features", help="Print features")
    parser.add_argument("--check", action="store_true", default=False,
                        dest="check_row_data", help="Check row data for NaN or float overflow")
    parser.add_argument("--bestfeatfile",
                        default="../../scripts/crystal/best_features_usefulness_clust.txt",
                        type=str,
                        dest="best_features_fname",
                        help="Name and position of best features file that lists the best features in order")

    # clustering
    parser.add_argument("--clusters", default=4, type=int,
                        dest="clusters", help="How many clusters to use")
    parser.add_argument("--scale", default=False, action="store_true",
                        dest="scale", help="Scale clustering")
    parser.add_argument("--nocomputed", default=True, action="store_false",
                        dest="computed", help="Don't add computed features")
    parser.add_argument("--samples", default=1000,
                        dest="samples_per_file", help="Samples per file")

    # number of configs to generate
    parser.add_argument("--confs", default="2-2", type=str,
                        dest="confs", help="Configs to generate")

    # code
    parser.add_argument("--basedir", type=str,
                        dest="basedir", help="The base directory of where the CryptoMiniSat source code is")

    options = parser.parse_args()
    prng = np.random.RandomState(options.seed)

    if options.basedir is None:
        print("ERROR: must set basedir")
        exit(-1)

    if options.fnames is None or len(options.fnames) == 0:
        print("ERROR: You must give the pandas file!")
        exit(-1)

    fnames = [f for f in options.fnames if "clust" not in f]
    print("Will add clustering to files: ")
    for f in fnames:
        print("->", f)

    if options.clusters <= 0:
        print("ERROR: You must give a '--clusters' option that is greater than 0")
        exit(-1)

    samples = None
    for f in fnames:
        print("===-- Sampling file %s --" % f)
        df = pd.read_pickle(f)
        print("options.samples_per_file:", options.samples_per_file)
        print("df.shape:", df.shape)
        new_samples = df.sample(options.samples_per_file, replace=True,
                                random_state=prng)
        samples = new_samples.append(samples)
        print("samples.shape:", samples.shape)
        del df

    if options.computed:
        helper.cldata_add_computed_features(samples, options.verbose)

    # clustering_setup is:
    #   feats_used, scaler, clust
    clustering_setup = [None, None]
    todo = [CLUST_TYPE_FILES, CLUST_TYPE_USEFULNESS]
    todo = [CLUST_TYPE_FILES]
    for clust_type in todo:
        c = Clustering(samples, clust_type)
        clustering_setup[clust_type] = c.cluster()
        name = get_cluster_name(clust_type)
        if termcolor_avail:
            cprint("===-- K-Means clustering **created ** type: %s --" % name,
               "green", "on_grey")
        else:
            cprint("===-- K-Means clustering **created ** type: %s --" % name)

    del samples

    for f in fnames:
        print("===-- Clustering file %s" % (f))
        df_orig = pd.read_pickle(f)
        df = df_orig.copy()
        if options.computed:
            helper.cldata_add_computed_features(df, options.verbose)

        for clust_type in todo:
            name = get_cluster_name(clust_type)
            print("===-- type: %s --" % name)
            df_orig[name] = get_clustering_label(df, *clustering_setup[clust_type])
            print("Distribution: \n%s" % df_orig[name].value_counts())

        cleanname = re.sub(r'\.dat$', '', f)
        dump_dataframe(df_orig, cleanname+"-clustered.dat")
        del df
        del df_orig



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function
import sqlite3
import optparse
import random

tables = ["tags", "timepassed", "memused"
          , "solverRun", "startup", "finishup"]

class Query:
    def __init__(self, dbfname):
        self.dbfname = dbfname
        self.conn = sqlite3.connect(self.dbfname)
        self.c = self.conn.cursor()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.conn.close()

    def add_ids(self):
        print("----------- adding IDs to db %s --------------" % self.dbfname)
        runid = random.randint(0, 2**32-1)

        query = """
        alter table `{tablename}` add column runid bigint(2) default null;
        update `{tablename}` set runid={runid};
        """

        for table in tables:
            print("adding runid {runid} to table {tablename}"
                      .format(runid=runid, tablename=table))
            for q in query.split("\n"):
                self.c.execute(q.format(runid=runid, tablename=table))

        print("Finished adding IDs to tables")
        self.c.execute("commit;")


    def create_empty_tables(self):
        query="""
        DROP TABLE IF EXISTS `tags`;
        CREATE TABLE `tags` (
          `name` varchar(500) NOT NULL,
          `val` varchar(500) NOT NULL
          , runid bigint(2) default null
        );

        DROP TABLE IF EXISTS `timepassed`;
        CREATE TABLE `timepassed` (
          `simplifications` bigint(20) NOT NULL,
          `conflicts` bigint(20) NOT NULL,
          `runtime` float NOT NULL,
          `name` varchar(200) NOT NULL,
          `elapsed` float NOT NULL,
          `timeout` int(20) DEFAULT NULL,
          `percenttimeremain` float DEFAULT NULL
          , runid bigint(2) default null
        );

        DROP TABLE IF EXISTS `memused`;
        CREATE TABLE `memused` (
          `simplifications` bigint(20) NOT NULL,
          `conflicts` bigint(20) NOT NULL,
          `runtime` float NOT NULL,
          `name` varchar(200) NOT NULL,
          `MB` int(20) NOT NULL
          , runid bigint(2) default null
        );

        DROP TABLE IF EXISTS `solverRun`;
        CREATE TABLE `solverRun` (
          `runtime` float NOT NULL,
          `gitrev` varchar(100) NOT NULL
          , runid bigint(2) default null
        );

        DROP TABLE IF EXISTS `startup`;
        CREATE TABLE `startup` (
          `startTime` datetime NOT NULL
          , runid bigint(2) default null
        );

        DROP TABLE IF EXISTS `finishup`;
        CREATE TABLE `finishup` (
          `endTime` datetime NOT NULL,
          `status` varchar(255) NOT NULL
          , runid bigint(2) default null
        );"""
        self.c.executescript(query)

    def merge_data(self, files):
        query = """
        attach '{fname}' as toMerge;
        BEGIN;
        insert into {table} select * from toMerge.{table};
        COMMIT;
        detach toMerge;
        """

        for f in files:
            print("Merging file %s" % f)
            for table in tables:
                print("-> Merging table %s" % table)
                self.c.executescript(query.format(fname=f, table=table))

if __name__ == "__main__":
    usage = "usage: %prog [options] sqlitedb"
    parser = optparse.OptionParser(usage=usage)
    parser.add_option("--verbose", "-v", action="store_true", default=False,
                      dest="verbose", help="Print more output")
    parser.add_option("--onlyconcat", action="store_true", default=False,
                      dest="onlyconcat", help="Only concatenate")

    (options, args) = parser.parse_args()

    if len(args) < 2:
        print("ERROR: You must give at least two arguments, the sqlite3 database files")
        exit(-1)

    if not options.onlyconcat:
        for fname in args:
            print("Using sqlite3db file %s" % fname)

            #peform queries
            with Query(fname) as q:
                q.add_ids()

        print("Finished adding IDs to all tables in all files")

    print("Merging tables...")
    with Query("merged.sqlitedb") as q:
        q.create_empty_tables()
        q.merge_data(args)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2017  Mate Soos
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.


import helper
import optparse
import time

class QueryFixPerc (helper.QueryHelper):
    def __init__(self, dbfname):
        super(QueryFixPerc, self).__init__(dbfname)

    def check_table_exists(self):
        q = "SELECT name FROM sqlite_master WHERE type='table' AND name='{table_name}';".format(
            table_name="used_later_percentiles_backup")
        self.c.execute(q)
        rows = self.c.fetchall()
        if len(rows) == 1:
            print("Already fixed")
            return True

        return False

    def copy_db(self):
        # Drop table
        q_drop = """
        DROP TABLE IF EXISTS `used_later_percentiles_backup`;
        """
        self.c.execute(q_drop)

        # Create and fill used_later_X tables
        q_create = """
        CREATE TABLE used_later_percentiles_backup AS
          SELECT *
          FROM used_later_percentiles;"""
        self.c.execute(q_create)

    def create_percentiles_table(self):
        # Drop table
        q_drop = """
        DROP TABLE IF EXISTS `used_later_percentiles`;
        """
        self.c.execute(q_drop)

        # Create and fill used_later_X tables
        q_create = """
        create table `used_later_percentiles` (
            `type_of_dat` string NOT NULL,
            `percentile_descr` string NOT NULL,
            `percentile` float DEFAULT NULL,
            `val` float NOT NULL
        );"""
        self.c.execute(q_create)

        idxs = """
        create index `used_later_percentiles_idx3` on `used_later_percentiles` (`type_of_dat`, `percentile_descr`, `percentile`, `val`);
        create index `used_later_percentiles_idx2` on `used_later_percentiles` (`type_of_dat`, `percentile_descr`, `val`);"""
        for q in idxs.split("\n"):
            self.c.execute(q)

    def fix(self):
        q2 = """
        insert into used_later_percentiles (type_of_dat, percentile_descr, percentile, val)
        {q}
        """

        t = time.time()
        q = "select * from used_later_percentiles_backup"
        self.c.execute(q)
        rows = self.c.fetchall()
        fixed = 0
        for i in range(len(rows)):
            if options.verbose:
                print("row: ", rows[i])
            row = rows[i]
            a = row[0][1:]
            c = None
            if row[1][:13] == "top_also_zero":
                b = "top_also_zero"
                c = row[1][13:]
            elif row[1][:12] == "top_non_zero":
                b = "top_non_zero"
                c = row[1][12:]
            elif row[1] == "avg":
                b = "avg"
                c = None
            else:
                print("c ? ", row[1])
                assert False, "What?"

            if c is not None:
                c = float(c.rstrip("_perc").strip("_"))
            else:
                c = "NULL"
            d = row[2]

            if a == "":
                # this was removed
                continue

            if options.verbose:
                print("a: ", a)
                print("b: ", b)
                print("c: ", c)
                print("d: ", d)
            self.c.execute(q2.format(q="select '{a}', '{b}', {c}, {d};".format(
                a=a, b=b, c=c, d=d)))
            fixed+=1

        print("Fixed:", fixed)

        # fix 100
        for name in ["short", "long", "forever"]:
            self.c.execute(q2.format(name=name, q="select '{name}', 'top_non_zero', 100.0, 0.0;".format(name=name)))
            self.c.execute(q2.format(name=name, q="select '{name}', 'top_also_zero', 100.0, 0.0;".format(name=name)))

        print("Fixed top 100 too")

if __name__ == "__main__":
    usage = "usage: %prog [options] file1.sqlite [file2.sqlite ...]"
    parser = optparse.OptionParser(usage=usage)

    # verbosity level
    parser.add_option("--verbose", "-v", action="store_true", default=False,
                      dest="verbose", help="Print more output")
    parser.add_option("--sql", action="store_true", default=False,
                      dest="dump_sql", help="Dump SQL queries")

    (options, args) = parser.parse_args()

    if len(args) < 1:
        print("ERROR: You must give at least one file")
        exit(-1)


    for f in args:
        print("Doing file:" , f)
        with QueryFixPerc(f) as q:
            if q.check_table_exists():
                continue
            q.copy_db()
            q.create_percentiles_table()
            q.fix()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2009-2020 Authors of CryptoMiniSat, see AUTHORS file
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from __future__ import print_function
import sqlite3
import argparse
import os
import struct
import time


class Query:
    def __init__(self, dbfname):
        self.conn = sqlite3.connect(dbfname)
        self.c = self.conn.cursor()

        self.cl_used = []
        self.cl_used_num = 0
        self.cl_used_total = 0

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.conn.commit()
        self.conn.close()

    def delete_tbls(self, table):
        queries = """
DROP TABLE IF EXISTS `{table}`;
create table `{table}` ( `clauseID` bigint(20) NOT NULL, `used_at` bigint(20) NOT NULL, `weight` float NOT NULL);
""".format(table=table)

        for l in queries.split('\n'):
            t2 = time.time()

            if options.verbose:
                print("Runnin query: ", l)
            self.c.execute(l)
            if options.verbose:
                print("Query T: %-3.2f s" % (time.time() - t2))

    def get_last_good(self, basefname):
        last_good = -1
        for i in range(100):
            tfname = "%s-%d" % (basefname, i)
            print("Checking if file %s exists" % tfname)
            if os.path.isfile(tfname):
              last_good = i

        if last_good == -1:
            print("ERROR: file does not exist at all")
            exit(-1)

        return last_good

    def add_used_clauses(self, usedClFname):
        last_good = self.get_last_good(usedClFname)
        tfname = "%s-%d" % (usedClFname, last_good)
        tfname_anc = tfname.replace("usedCls", "usedCls-anc")


        for fname,table in [(tfname, "used_clauses"), (tfname_anc, "used_clauses_anc")]:
            t = time.time()
            print("Adding data from file %s to DB %s" % (fname, table))

            self.cl_used = []
            self.cl_used_num = 0
            self.cl_used_total = 0
            with open(fname, "rb") as f:
                while True:
                    b1 = f.read(8)
                    if not b1:
                        # end of file
                        break
                    b2 = f.read(8)
                    cl_id = struct.unpack("<q", b1)[0]
                    conf = struct.unpack("<q", b2)[0]

                    weight = 1.0
                    if "used_clauses_anc" in table:
                        b3 = f.read(4)
                        weight = struct.unpack("<f", b3)[0]

                    dat = (cl_id, conf, weight)
                    if options.verbose:
                        print("Use:", dat)
                    self.cl_used.append(dat)
                    self.cl_used_num += 1
                    self.cl_used_total += 1
                    if self.cl_used_num > 10000:
                        self.dump_used_clauses(table)

            self.dump_used_clauses(table)
            print("Added use data to table %s from file %s: %d time: T: %-3.2f s" %
                  (table, fname, self.cl_used_total, time.time() - t))

    def dump_used_clauses(self, table):
        self.c.executemany("""
        INSERT INTO %s (
        `clauseID`,
        `used_at`,
        `weight`)
        VALUES (?, ?, ?);""" % table, self.cl_used)
        self.cl_used = []
        self.cl_used_num = 0


if __name__ == "__main__":
    usage = """usage: %(prog)s [options] sqlite_db usedCls

Adds used_clauses to the SQLite database"""

    parser = argparse.ArgumentParser(usage=usage)


    parser.add_argument("sqlitedb", type=str, metavar='USEDCLS')
    parser.add_argument("usedcls", type=str, metavar='USEDCLS')
    parser.add_argument("--verbose", "-v", action="store_true", default=False,
                      dest="verbose", help="Print more output")

    options = parser.parse_args()

    if options.sqlitedb is None:
        print("Error you must give an SQLite DB. Please follow usage.")
        print(usage)
        exit(-1)

    if options.usedcls is None:
        print("Error you must give an Used CLS list. Please follow usage.")
        print(usage)
        exit(-1)

    print("Using sqlite3db file %s" % options.sqlitedb)
    print("Base used_clauses file is %s" % options.usedcls)

    with Query(options.sqlitedb) as q:
        q.delete_tbls("used_clauses")
        q.delete_tbls("used_clauses_anc")
        q.add_used_clauses(options.usedcls)

    print("Finished adding good lemma indicators to db %s" % options.sqlitedb)

    exit(0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2009-2020 Authors of CryptoMiniSat, see AUTHORS file
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import ast

BINOP_SYMBOLS = {}
BINOP_SYMBOLS[ast.Add] = '+'
BINOP_SYMBOLS[ast.Sub] = '-'
BINOP_SYMBOLS[ast.Mult] = '*'
BINOP_SYMBOLS[ast.Div] = '/'
BINOP_SYMBOLS[ast.Mod] = '%'
BINOP_SYMBOLS[ast.Pow] = '**'
BINOP_SYMBOLS[ast.LShift] = '<<'
BINOP_SYMBOLS[ast.RShift] = '>>'
BINOP_SYMBOLS[ast.BitOr] = '|'
BINOP_SYMBOLS[ast.BitXor] = '^'
BINOP_SYMBOLS[ast.BitAnd] = '&'
BINOP_SYMBOLS[ast.FloorDiv] = '//'

BOOLOP_SYMBOLS = {}
BOOLOP_SYMBOLS[ast.And] = 'and'
BOOLOP_SYMBOLS[ast.Or] = 'or'

CMPOP_SYMBOLS = {}
CMPOP_SYMBOLS[ast.Eq] = '=='
CMPOP_SYMBOLS[ast.NotEq] = '!='
CMPOP_SYMBOLS[ast.Lt] = '<'
CMPOP_SYMBOLS[ast.LtE] = '<='
CMPOP_SYMBOLS[ast.Gt] = '>'
CMPOP_SYMBOLS[ast.GtE] = '>='
CMPOP_SYMBOLS[ast.Is] = 'is'
CMPOP_SYMBOLS[ast.IsNot] = 'is not'
CMPOP_SYMBOLS[ast.In] = 'in'
CMPOP_SYMBOLS[ast.NotIn] = 'not in'

class ccg:
    def fix_feat_name(x):
        return "df[\"%s\"]" % x


    def to_source(node, updater=fix_feat_name):
        """This function can convert a node tree back into python sourcecode.
        This is useful for debugging purposes, especially if you're dealing with
        custom asts not generated by python itself.

        It could be that the sourcecode is evaluable when the AST itself is not
        compilable / evaluable.  The reason for this is that the AST contains some
        more data than regular sourcecode does, which is dropped during
        conversion.
        """
        generator = ccg.SourceGenerator()
        generator.update=updater
        generator.visit(node)

        return ''.join(generator.result)

    class SourceGenerator(ast.NodeVisitor):
        """This visitor is able to transform a well formed syntax tree into python
        sourcecode.  For more details have a look at the docstring of the
        `node_to_source` function.
        """

        def __init__(self):
            self.result = []

        def write(self, x):
            self.result.append(x)

        def body(self, statements):
            for stmt in statements:
                self.visit(stmt)

        def body_or_else(self, node):
            self.body(node.body)
            if node.orelse:
                self.write('else:')
                self.body(node.orelse)

        def signature(self, node):
            want_comma = []
            def write_comma():
                if want_comma:
                    self.write(', ')
                else:
                    want_comma.append(True)

            padding = [None] * (len(node.args) - len(node.defaults))
            for arg, default in zip(node.args, padding + node.defaults):
                write_comma()
                self.visit(arg)
                if default is not None:
                    self.write('=')
                    self.visit(default)
            if node.vararg is not None:
                write_comma()
                self.write('*' + node.vararg)
            if node.kwarg is not None:
                write_comma()
                self.write('**' + node.kwarg)

        # Statements

        def visit_Expr(self, node):
            self.write("(")
            self.generic_visit(node)
            self.write(")")

        # Expressions

        def visit_Attribute(self, node):
            x = self.update(node.value.id + '.' + node.attr)
            self.write(x)


        # visiting functions. only log2
        # should be: df[divisor].apply(np.log2)
        def visit_Call(self, node):
            if node.func.id != "log2":
                self.visit(node.func)
            self.write('(')

            for arg in node.args:
                self.visit(arg)
            for keyword in node.keywords:
                self.write(keyword.arg + '=')
                self.visit(keyword.value)

            self.write(".apply(np.log2)")
            self.write(')')


        def visit_Name(self, node):
            #print("type:", type(node.id))
            if node.id != "log2":
                x = self.update(node.id)
            else:
                x = node.id

            self.write(x)

        def visit_Str(self, node):
            self.write(repr(node.s))

        def visit_Bytes(self, node):
            self.write(repr(node.s))

        def visit_Num(self, node):
            self.write(repr(node.n))

        def visit_BinOp(self, node):
            self.write('(')
            self.visit(node.left)
            self.write(' %s ' % BINOP_SYMBOLS[type(node.op)])
            self.visit(node.right)
            self.write(')')

        def visit_BoolOp(self, node):
            self.write('(')
            for idx, value in enumerate(node.values):
                if idx:
                    self.write(' %s ' % BOOLOP_SYMBOLS[type(node.op)])
                self.visit(value)
            self.write(')')

        def visit_Compare(self, node):
            self.write('(')
            self.visit(node.left)
            for op, right in zip(node.ops, node.comparators):
                self.write(' %s ' % CMPOP_SYMBOLS[type(op)])
                self.visit(right)
            self.write(')')

        # Helper Nodes

        def visit_alias(self, node):
            self.write(node.name)
            if node.asname is not None:
                self.write(' as ' + node.asname)

        def visit_arguments(self, node):
            self.signature(node)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2009-2020 Authors of CryptoMiniSat, see AUTHORS file
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import collections
import sys

toaddheader = []
toaddval = []
for x in range(1, len(sys.argv)):
    if x % 2 == 1:
        toaddheader.append(sys.argv[x])
    else:
        toaddval.append(sys.argv[x])


infilenames = []
infilenames.append("signals.csv")
infilenames.append("solveTimes_rev.csv")

# main data
files = {}


###############
# Fill file names
###############
allfiles = open("allFiles.csv", "r")
for l in allfiles:
    l = l.strip()
    files[l] = {"SAT":"?"}

###############
# Fill times, signals
###############
infiles = []
for x in infilenames:
    infiles.append(open(x, "r"))
for f,fname in zip(infiles, infilenames):
    for l in f:
        l = l.strip()
        l = l.split(" ")

        # print("appending to %s : %s" % (l[0], l[1]))
        files[l[0]][fname] = l[1]


###############
# Fill SAT/UNSAT
###############
with open("solvedUNSAT.csv") as f:
    for l in f:
        l = l.strip()
        files[l]["SAT"] = "UNSAT"

with open("solvedSAT.csv") as f:
    for l in f:
        l = l.strip()
        files[l]["SAT"] = "SAT"


###############
# Print header
###############
toprint = "fname,"
for x in range(len(infilenames)):
    toprint += infilenames[x].replace(".csv", "")
    if x+1 < len(infilenames):
        toprint += ","

for x in toaddheader:
    toprint+=","+x
toprint+=",SAT"

print(toprint)



###############
# Print lines, in order
###############
od = collections.OrderedDict(sorted(files.items()))
for k, v in od.items():
    toprint = ""
    for fname in infilenames+["SAT"]:
        if fname in v:
            toprint += "%s," % v[fname]
        else:
            toprint += "?,"

    for x in toaddval:
        toprint+= x + ","

    toprint = toprint.rstrip(",")
    print("%s,%s" % (k, toprint))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2009-2020 Authors of CryptoMiniSat, see AUTHORS file
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from __future__ import print_function
import sqlite3
import argparse
import time
import pickle
import re
import pandas as pd
import numpy as np
import os.path
import sys
import helper


def dump_df(df):
    # rename columns if we have to:
    cleanname = re.sub(r'\.cnf.gz.sqlite$', '', options.fname)
    cleanname = re.sub(r'\.db$', '', options.fname)
    cleanname = re.sub(r'\.sqlitedb$', '', options.fname)
    cleanname = "{cleanname}-vardata".format(
        cleanname=cleanname)
    dump_dataframe(df, cleanname)


def dump_dataframe(df, name):
    if options.dump_csv:
        fname = "%s.csv" % name
        print("Dumping CSV data to:", fname)
        df.to_csv(fname, index=False, columns=sorted(list(df)))

    fname = "%s.dat" % name
    print("Dumping pandas data to:", fname)
    with open(fname, "wb") as f:
        pickle.dump(df, f)


class QueryVar (helper.QueryHelper):
    def __init__(self, dbfname):
        super(QueryVar, self).__init__(dbfname)

    def create_indexes(self):
        helper.drop_idxs(self.c)

        print("Recreating indexes...")
        t = time.time()
        queries = """
        create index `idxclid8` on `var_data_picktime` ( `var`, `sumConflicts_at_picktime`, `latest_vardist_feature_calc`);
        create index `idxclid81` on `var_data_picktime` ( `var`, `sumConflicts_at_picktime`);
        create index `idxclid82` on `var_dist` ( `var`, `latest_vardist_feature_calc`);
        create index `idxclid9` on `var_data_fintime` ( `var`, `sumConflicts_at_picktime`);

        create index `idxclid10` on `dec_var_clid` ( `var`, `sumConflicts_at_picktime`, `clauseID`);

        create index `idxclid-s2` on `restart_dat_for_var` (`conflicts`, `latest_satzilla_feature_calc`, `branch_strategy`);
        create index `idxclid-s4` on `satzilla_features` (`latest_satzilla_feature_calc`);

        create index `idxclid-s1` on `sum_cl_use` ( `clauseID`, `num_used`);
        """

        for l in queries.split('\n'):
            t2 = time.time()

            if options.verbose:
                print("Creating/dropping index: ", l)
            self.c.execute(l)
            if options.verbose:
                print("Index dropping&creation T: %-3.2f s" %
                      (time.time() - t2))

        print("indexes dropped&created T: %-3.2f s" % (time.time() - t))

    def fill_var_data_use(self):
        print("Filling var data use...")

        t = time.time()
        q = "delete from `var_data_use`;"
        self.c.execute(q)
        print("var_data_use cleared T: %-3.2f s" % (time.time() - t))

        t = time.time()
        q = """
        insert into var_data_use

        (`var`
        , `sumConflicts_at_picktime`

        , `cls_marked`
        , `useful_clauses_used`)

        select
        dvclid.var
        , dvclid.sumConflicts_at_picktime

        -- measures for good
        , count(cls.num_used) as cls_marked
        , sum(cls.num_used) as useful_clauses_used

        FROM dec_var_clid as dvclid join sum_cl_use as cls
        on cls.clauseID = dvclid.clauseID

        -- avoid division by zero below

        group by dvclid.var, dvclid.sumConflicts_at_picktime
        ;
        """
        if options.verbose:
            print("query:", q)
        self.c.execute(q)

        print("var_data_use filled T: %-3.2f s" % (time.time() - t))

        t = time.time()
        q = """
        UPDATE var_data_use SET useful_clauses_used = 0
        WHERE useful_clauses_used IS NULL
        """
        self.c.execute(q)

        print("var_data_use updated T: %-3.2f s" % (time.time() - t))

    def create_vardata_df(self, min_val, max_val, branch_str = None):
        not_cols = [
            "clid_start_incl"
            , "clid_end_notincl"
            , "decided_pos"
            , "propagated_pos"
            , "restarts"
            , "var"
            , "cls_marked"]
        var_data_picktime = helper.query_fragment(
            "var_data_picktime", not_cols, "var_data_picktime", options.verbose, self.c)
        var_data_fintime = helper.query_fragment(
            "var_data_fintime", not_cols, "var_data_fintime", options.verbose, self.c)

        not_cols =[
            "var"
            , "latest_vardist_feature_calc"
            , "conflicts"
        ]
        var_dist = helper.query_fragment(
            "var_dist", not_cols, "var_dist", options.verbose, self.c)

        not_cols =[
            "simplifications"
            , "restarts"
            , "conflicts"
            , "latest_satzilla_feature_calc"
            , "runtime"
            , "propagations"
            , "decisions"
            , "flipped"
            , "replaced"
            , "eliminated"
            , "set"]
        rst = helper.query_fragment(
            "restart_dat_for_var", not_cols, "rst", options.verbose, self.c)

        not_cols =[
            ""]
        szfeat = helper.query_fragment(
            "satzilla_features", not_cols, "szfeat", options.verbose, self.c)

        not_cols =[
            "useful_clauses"
            , "sumConflicts_at_picktime"
            , "var"]
        var_data_use = helper.query_fragment(
            "var_data_use", not_cols, "var_data_use", options.verbose, self.c)

        not_cols =[
            "clauseID"
            , "first_confl_used"
            , "last_confl_used"]
        sum_cl_use = helper.query_fragment(
            "sum_cl_use", not_cols, "sum_cl_use", options.verbose, self.c)


        my_branch_str = ""
        if branch_str is not None:
            my_branch_str = "and rst.branch_strategy = {branch_str}".format(
                branch_str=branch_str)

        q = """
        select
        sum_cl_use.num_used as `x.num_used`
        {rst}
        {var_dist}
        {var_data_picktime}
        {var_data_fintime}
        {szfeat}
        {sum_cl_use}

        FROM
        var_data_picktime
        , var_data_fintime
        , sum_cl_use
        , dec_var_clid
        , var_dist
        , restart_dat_for_var as rst
        , satzilla_features as szfeat

        WHERE
        var_data_picktime.sumConflicts_at_picktime > 15000
        and sum_cl_use.clauseID = dec_var_clid.clauseID

        and var_data_picktime.var = dec_var_clid.var
        and var_data_picktime.sumConflicts_at_picktime = dec_var_clid.sumConflicts_at_picktime

        and var_data_fintime.var = dec_var_clid.var
        and var_data_fintime.sumConflicts_at_picktime = dec_var_clid.sumConflicts_at_picktime

        and var_dist.var = dec_var_clid.var
        and var_dist.latest_vardist_feature_calc = var_data_picktime.latest_vardist_feature_calc

        and rst.conflicts = var_data_picktime.sumConflicts_at_picktime
        and rst.latest_satzilla_feature_calc = szfeat.latest_satzilla_feature_calc

        and sum_cl_use.num_used >= {min_val}
        and sum_cl_use.num_used <= {max_val}

        {my_branch_str}

        order by random()
        limit {limit}
        """.format(
            rst=rst,
            var_data_use=var_data_use,
            var_data_picktime=var_data_picktime,
            var_data_fintime=var_data_fintime,
            var_dist=var_dist,
            sum_cl_use=sum_cl_use,
            szfeat=szfeat,
            limit=options.limit,
            min_val=min_val,
            max_val=max_val,
            my_branch_str=my_branch_str,
            min_cls_below=options.min_cls_below)

        df = pd.read_sql_query(q, self.conn)
        print("DF dimensions:", df.shape)
        return df


if __name__ == "__main__":
    usage = "usage: %(prog)s [options] file.sqlite"
    parser = argparse.ArgumentParser(usage=usage)

    parser.add_argument("fname", type=str, metavar='SQLITEFILE')
    parser.add_argument("--verbose", "-v", action="store_true",
                        default=False, dest="verbose", help="Print more output")
    parser.add_argument("--csv", action="store_true", default=False,
                        dest="dump_csv", help="Dump CSV (for weka)")
    parser.add_argument("--limit", type=int, default=10000,
                        dest="limit", help="How many data points")
    parser.add_argument("--minclsbelow", type=int, default=1,
                        dest="min_cls_below", help="Minimum number of clauses below to generate data point")
    parser.add_argument("--unbalanced", action="store_true",
                        default=False, dest="unbalanced", help="Get unbalanced output")

    options = parser.parse_args()

    if options.fname is None:
        print("ERROR: You must give exactly one file")
        exit(-1)

    np.random.seed(2097483)
    with QueryVar(options.fname) as q:
        q.create_indexes()
        q.fill_var_data_use()

        if not options.unbalanced:
            dfs = []
            for branch_str in range(4):
                print("Doing == 0 use:")
                dfs.append(q.create_vardata_df(0, 0, branch_str))
                print("Doing >0 use:")
                dfs.append(q.create_vardata_df(1, 200000, branch_str))
                print("Finished branch_str: ", branch_str)
            df_full = pd.concat(dfs, sort=False)
        else:
            df_full = q.create_vardata_df(0, 2000000000)

        print("Final DF dimensions:", df_full.shape)

        dump_df(df_full)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2022 Authors of CryptoMiniSat, see AUTHORS file
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 2
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import sys
import subprocess

def parse_lit(x):
    inv = 1
    if x[0] == '~':
        inv = -1
        x = x[1:]

    assert x[0] == 'x'
    v = int(x[1:])
    return v*inv

def leq_translate(c):
    #if len(c.lhs) == 1 and c.lhs[0][0] == 1 and c.cutoff == 0:
        #print("%d %d 0" % (-c.lhs[0][1], c.rhs))
        #print("%d %d 0" % (c.lhs[0][1], -c.rhs))
        #return
    #if len(c.lhs) == 1 and c.lhs[0][0] == -1 and c.cutoff == 1:
        #print("%d %d 0" % (-c.lhs[0][1], c.rhs))
        #print("%d %d 0" % (c.lhs[0][1], -c.rhs))
        #return

    #if len(c.lhs) == 1 and c.rhs == None and c.cutoff == 1:
        #if c.lhs[0][0] == 1:
            #print("%d 0" % c.lhs[0][1])
            #return
        #sys.stderr.write("ERRROOOR\n")
        #exit(-1)

    ins = []
    for l in c.lhs:
        if l[0] == 1:
            ins.append(l[1])
            continue
        else:
            ins.append(-l[1])
            c.cutoff+=1

    pr = "b "
    for l in ins:
        pr += "%d " % l

    pr += "0 "
    pr += "%d " % c.cutoff

    if c.rhs is not None:
        pr += "%d " % c.rhs
    print("%s" % pr)

class Constr:
    def __init__(self, lhs, cutoff, rhs, sign):
        self.lhs = lhs
        self.cutoff = cutoff
        self.rhs = rhs
        self.sign = sign

    def __repr__(self):
        return "lhs: %s cutoff: %s sign: %s rhs: %s" % (self.lhs, self.cutoff, self.sign, self.rhs)

    def simple(self):
        if self.sign != "=" and self.sign != ">=":
            return

        for l in self.lhs:
            if l[0] == 1 or l[0] == -1:
                continue
            else:
                return False

        return True

    def translate(self):
        print("c Doing simple:", self)
        assert self.simple()
        if self.sign == ">=":
            leq_translate(self)
        elif self.sign == "=":
            leq_translate(self)
            c = Constr(self.lhs, self.cutoff, self.rhs, self.sign)
            c.cutoff = len(c.lhs) - c.cutoff
            for i in range(len(c.lhs)):
                c.lhs[i][0] = -c.lhs[i][0]
            leq_translate(c)
        else:
            print("ERRROR")
            exit(-1)





def parse_constr(line):
    lhs = []
    cutoff = None
    rhs = None
    sign = ""

    at_lhs = True
    at_cutoff = False
    at_rhs = False
    done = False
    line = line.split(' ')
    i = 0
    #print("line:", line)
    while i < len(line):
        #print("SO FAR -- lhs: ", lhs, "cutoff: ", cutoff, "rhs: ", rhs, "at_lhs:", at_lhs, "at_cutoff:", at_cutoff, "at_rhs:", at_rhs)

        k = line[i].strip()
        if k == '':
            i+=1
            continue

        if at_lhs:
            if k == ">=" or k == "=":
                sign = k
                at_lhs = False
                at_cutoff= True
                i+=1
                continue
            #else:
                #sys.stderr.write("ERRROOOR: input has:" + k)
                #exit(-1)
            coeff = int(k)
            assert i+1 < len(line)
            i+=1
            k = line[i].strip()
            lit = parse_lit(k)
            lhs.append([coeff, lit])
            i+=1
            continue

        if at_cutoff:
            if k == "<->":
                at_cutoff= False
                at_rhs = True
                i+=1
                continue
            if k == ";":
                done = True
                break
            cutoff = int(k)
            i+=1
            continue

        if at_rhs:
            if k == ";":
                done = True
                break
            assert k[0] == 'x'
            rhs = parse_lit(k)
            i+=1
            continue

        assert False

    assert done

    #if rhs is not None:
        #print("lhs: ", lhs, "cutoff: ", cutoff, "rhs: ", rhs)
    #else:
        #print("lhs: ", lhs, "cutoff: ", cutoff)

    return Constr(lhs, cutoff, rhs, sign)


def do_pbsugar(constr, maxvar):
    with open("sugarin", "w") as f:
        f.write("* #variable= %d #constraint= 1\n" % maxvar)
        for i in constr.lhs:
            f.write("%d x%d " % (i[0],i[1]))
        f.write(" >= %d ;\n" % constr.cutoff)
    assert constr.rhs is None

    args = ("./pbsugar", "-map", "x", "-n", "sugarin", "-sat", "sugarout")
    popen = subprocess.Popen(args, stdout=subprocess.PIPE)
    popen.wait()
    output = popen.stdout.read()
    #print("c pbsugar output: ", output)

    # get map
    varmap_end = 0
    varmap = {}
    with open("x", "r") as f:
        first = True
        for line in f:
            if first:
                num_max = int(line.strip())
                assert num_max == maxvar # or it had to use some extra vars
                first = False
                continue
            line=line.strip().split(' ')
            assert line[0][0] == 'x'
            new = int(line[1])
            orig = int(line[0][1:])
            varmap[new] = orig
            if new > varmap_end:
                varmap_end = new

    towrite = ""
    with open("sugarout", "r") as f:
        for line in f:
            if "p cnf" in line:
                continue
            line = line.strip().split()
            for lit in line:
                lit = int(lit)
                sign = lit < 0
                var = abs(lit)
                if var == 0:
                    towrite+= '0'
                    print(towrite)
                    towrite = ""
                    continue

                if var in varmap:
                    # part of original formula
                    var2 = varmap[var]
                else:
                    # fresh variable by pbsugar
                    assert var > varmap_end
                    maxvar+=1
                    var2 = maxvar

                if sign:
                    lit2 = -var2
                else:
                    lit2 = var2
                towrite+="%d " % lit2

    return maxvar


if __name__ == "__main__":
    if len(sys.argv) != 2:
        print("ERROR: must give ONE input, the OPB file")
        exit(-1)

    fname = sys.argv[1]

    # parse input
    ind = []
    constr = []
    maxvar = None
    with open(fname, "r") as f:
        # parse ind
        for line in f:
            line = line.strip()
            if len(line) == 0:
                continue

            if "#variable" in line:
                maxvar = int(line.split()[2])
                print("c maxvar: ", maxvar)
                continue

            if "* ind" in line:
                line = line[5:]
                for k in line.split(' '):
                    k = k.strip()
                    if k != "0" and k != '':
                        ind.append(int(k))
                continue

            if line[0] == '*':
                continue

            constr.append(parse_constr(line))

    # deal with independent
    pr = "c ind "
    for i in ind:
        pr += "%d " % i
    print("%s 0" % pr)

    # create output
    for c in constr:
        print("c doing constraint: ", c)
        if c.simple():
            c.translate()
        else:
            print("c Using pbsugar to deal with: ", c)
            maxvar = do_pbsugar(c, maxvar)


